# pylint: disable = line-too-long, invalid-name, missing-docstring
""" Tests for `typing_json.dump`, `typing_json.dumps`, `typing_json.load` and `typing_json.loads`. """

# standard imports
import io
from decimal import Decimal
from typing import Union

# external dependencies
from typing_extensions import Literal

# internal imports
from typing_json import dump, dumps, load, loads


BASETYPES_VALUES = [
    (True, bool),
    (1, int),
    (1.5, float),
    ("hello", str),
    (None, type(None)),
    (None, None),
    (True, Literal[True, 1.5, None]),
    (1.5, Literal[True, 1.5, None]),
    (None, Literal[True, 1.5, None]),
    (True, Union[bool, float, Decimal]),
    (1.5, Union[bool, float, Decimal]),
]


def test_basetypes():
    for val, t in BASETYPES_VALUES:
        assert val == loads(dumps(val, t), t)
        with io.StringIO() as f:
            dump(val, t, f)
            data = f.getvalue()
        with io.StringIO(data) as f:
            assert load(f, t) == val

def test_error():
    try:
        with io.StringIO() as f:
            dump(b"hello", bytes, f)
        assert False
    except TypeError:
        assert True
    try:
        dumps(b"hello", bytes)
        assert False
    except TypeError:
        assert True
    try:
        with io.StringIO() as f:
            load(f, bytes)
        assert False
    except TypeError:
        assert True
    try:
        loads("hello", bytes)
        assert False
    except TypeError:
        assert True



""" Tests for `typing_json.typechecking.is_json_encodable` """
# pylint: disable = line-too-long, invalid-name, missing-docstring

# standard imports
from typing import Union, Optional, List, Tuple, Set, FrozenSet, Mapping, Dict, OrderedDict, NamedTuple
from decimal import Decimal

# external dependencies
from typing_extensions import Literal

# internal imports
from typing_json.typechecking import TYPECHECKABLE_BASE_TYPES
from typing_json.encoding import is_json_encodable, JSON_BASE_TYPES


NONENCODABLE_BASE_TYPES = tuple(t for t in TYPECHECKABLE_BASE_TYPES if t not in JSON_BASE_TYPES+(Decimal,))

def failure_callback(message: str) -> None:
    print(message)


def test_is_json_encodable_basetypes():
    """
        Tests that base types which are meant
        to be JSON encodable are actually JSON encodable.
    """
    for t in JSON_BASE_TYPES:
        assert is_json_encodable(t, failure_callback=failure_callback)
    assert is_json_encodable(None, failure_callback=failure_callback)
    assert is_json_encodable(Decimal, failure_callback=failure_callback)

def test_is_json_encodable_classes():
    """
        Tests that some types which are not meant
        to be JSON encodable are actually not JSON encodable.
    """
    class t:
        # pylint: disable = all
        name: str
        value: int
    assert not is_json_encodable(t, failure_callback=failure_callback)


def test_is_json_encodable_enums():
    """
        Tests that enums are JSON encodable.
    """
    from enum import Enum
    class Color(Enum):
        # pylint:disable=all
        Red = 0
        Green = 1
        Blue = 2
    assert is_json_encodable(Color, failure_callback=failure_callback)


def test_is_json_encodable_collections1():
    """
        Tests that collections of JSON base types, `decimal.Decimal`
        and enums are typecheckable.
    """
    from enum import Enum
    class Color(Enum):
        # pylint:disable=all
        Red = 0
        Green = 1
        Blue = 2
    for t in JSON_BASE_TYPES+(Decimal, Color):
        assert is_json_encodable(FrozenSet[t], failure_callback=failure_callback)
        assert is_json_encodable(Union[t, Tuple[t]], failure_callback=failure_callback)
        assert is_json_encodable(Optional[t], failure_callback=failure_callback)
        assert is_json_encodable(Tuple[t, ...], failure_callback=failure_callback)
        assert is_json_encodable(Tuple[t], failure_callback=failure_callback)
        assert is_json_encodable(Tuple[t, int], failure_callback=failure_callback)
        assert is_json_encodable(Tuple[t, int, str], failure_callback=failure_callback)
        assert is_json_encodable(Set[t], failure_callback=failure_callback)
        assert is_json_encodable(List[t], failure_callback=failure_callback)
        assert is_json_encodable(Dict[str, t], failure_callback=failure_callback)
        assert is_json_encodable(Dict[t, int], failure_callback=failure_callback)
        assert is_json_encodable(OrderedDict[str, t], failure_callback=failure_callback)
        assert is_json_encodable(OrderedDict[t, int], failure_callback=failure_callback)
        assert is_json_encodable(Mapping[str, t], failure_callback=failure_callback)
        assert is_json_encodable(Mapping[t, int], failure_callback=failure_callback)


def test_is_json_encodable_collections2():
    """
        Tests that collections of non-typecheckable
        types are not typecheckable.
    """
    for t in NONENCODABLE_BASE_TYPES:
        assert not is_json_encodable(FrozenSet[t], failure_callback=failure_callback)
        assert not is_json_encodable(Union[t, Tuple[int]], failure_callback=failure_callback)
        assert not is_json_encodable(Optional[t], failure_callback=failure_callback)
        assert not is_json_encodable(Tuple[t, ...], failure_callback=failure_callback)
        assert not is_json_encodable(Tuple[t], failure_callback=failure_callback)
        assert not is_json_encodable(Tuple[t, int], failure_callback=failure_callback)
        assert not is_json_encodable(Set[t], failure_callback=failure_callback)
        assert not is_json_encodable(List[t], failure_callback=failure_callback)
        assert not is_json_encodable(Dict[str, t], failure_callback=failure_callback)
        assert not is_json_encodable(Dict[t, int], failure_callback=failure_callback)
        assert not is_json_encodable(OrderedDict[str, t], failure_callback=failure_callback)
        assert not is_json_encodable(OrderedDict[t, int], failure_callback=failure_callback)
        assert not is_json_encodable(Mapping[str, t], failure_callback=failure_callback)
        assert not is_json_encodable(Mapping[t, int], failure_callback=failure_callback)



def test_is_json_encodable_literal():
    """
        Tests that Literal types are JSON encodable if and only if all
        literals are of JSON base type or `decimal.Decimal`.
    """
    assert is_json_encodable(Literal["hello", 1, 2.5, None], failure_callback=failure_callback)
    assert not is_json_encodable(Literal["hello", 1, 2.5, None, 1+1j], failure_callback=failure_callback)

def test_is_json_encodable_namedtuple():
    """
        Tests that namedtuples are JSON encodable if
        and only if their attributes are of JSON encodable type.
    """
    class HashableNamedTupleExampleT(NamedTuple):
        name: Tuple[str, ...]
        value: Union[int, float]
    assert is_json_encodable(HashableNamedTupleExampleT, failure_callback=failure_callback)
    class NonHashableNamedTupleExampleT(NamedTuple):
        name: List[str]
        value: Union[int, float, complex]
    assert not is_json_encodable(NonHashableNamedTupleExampleT, failure_callback=failure_callback)
    class NonTypechekableT:
        name: str
        val: int
    class NonTypecheckableNamedTupleExampleT(NamedTuple):
        name: List[NonTypechekableT]
        value: Union[int, float]
    assert not is_json_encodable(NonTypecheckableNamedTupleExampleT, failure_callback=failure_callback)



""" Tests for `typing_json.typechecking.to_json_obj` """
# pylint: disable = line-too-long, invalid-name, missing-docstring

# standard imports
import json
import typing
from typing import Union, Optional, List, Tuple, Set, FrozenSet, Mapping, Dict, NamedTuple, Deque
from decimal import Decimal
from collections import deque, OrderedDict

# external dependencies
from typing_extensions import Literal

# internal imports
from typing_json.typechecking import TYPECHECKABLE_BASE_TYPES
from typing_json.encoding import to_json_obj, JSON_BASE_TYPES
from typing_json.decoding import from_json_obj, JSON_BASE_TYPES

# (bool, int, float, str, type(None))
NONENCODABLE_BASE_TYPES = tuple(t for t in TYPECHECKABLE_BASE_TYPES if t not in JSON_BASE_TYPES+(Decimal,))

def failure_callback(message: str) -> None:
    print(message)


def test_from_json_obj_typeerrors():
    """ Checks that `from_json_obj` raises `TypeError` on non-typecheckable types or non-json-encodable types.
    """
    class NonTypechekableT:
        name: str
        val: int
        def __init__(self, name, val):
            self.name = name
            self.val = val
    x = NonTypechekableT("x", 0)
    try:
        from_json_obj(x, NonTypechekableT)
        assert False
    except TypeError:
        assert True


BASETYPES_ENCODINGS = [
    (True, bool, True),
    (1, int, 1),
    (1.5, float, 1.5),
    ("hello", str, "hello"),
    (None, type(None), None),
    (None, None, None),
    (Decimal("1.0"), Decimal, 1),
    (Decimal("1.5"), Decimal, "1.5"),
    (True, Literal[True, 1.5, None], True),
    (1.5, Literal[True, 1.5, None], 1.5),
    (None, Literal[True, 1.5, None], None),
    (True, Union[bool, float, Decimal], True),
    (1.5, Union[bool, float, Decimal], 1.5),
    (Decimal("1.0"), Union[bool, None, Decimal], 1),
    (Decimal("1.5"), Union[bool, None, Decimal], "1.5"),
    (1, int, Decimal(1)),
    (1, int, Decimal("1.0")),
]


def test_from_json_obj_basetypes():
    """ Test `to_json_obj` on JSON base types and `decimal.Decimal`. """
    for val, t, encoding in BASETYPES_ENCODINGS:
        assert from_json_obj(encoding, t) == val

def test_from_json_obj_numbers():
    assert from_json_obj(1.5, Decimal) == Decimal(1.5)
    assert from_json_obj(1, float) == 1.0
    assert from_json_obj(Decimal("1.5"), float) == 1.5
    assert from_json_obj(Decimal("1.0"), float) == 1.0
    try:
        from_json_obj(Decimal("1.5"), float, cast_decimal=False)
        assert False
    except TypeError:
        assert True


from enum import Enum
class EnumT(Enum):
    # pylint:disable=all
    Red = 0
    Green = 1
    Blue = 2


def test_from_json_obj_enums():
    assert from_json_obj("Red", EnumT) == EnumT.Red


def test_from_json_obj_lists():
    """ Checks that lists are decoded correctly. """
    encoding = [1, 2, 3]
    val = from_json_obj(encoding, List[int])
    assert isinstance(val, list)
    assert len(val) == 3
    assert all(val[i] == encoding[i] for i in range(len(encoding)))


def test_from_json_obj_tuples1():
    """ Checks that fixed-length tuples are decoded correctly. """
    encoding = [1, 2, 3]
    val = from_json_obj(encoding, Tuple[int, int, int])
    assert isinstance(val, tuple)
    assert len(val) == 3
    assert all(val[i] == encoding[i] for i in range(len(encoding)))


def test_from_json_obj_tuples2():
    """ Checks that variadic tuples are decoded correctly. """
    encoding = [1, 2, 3]
    val = from_json_obj(encoding, Tuple[int, ...])
    assert isinstance(val, tuple)
    assert len(val) == 3
    assert all(val[i] == encoding[i] for i in range(len(encoding)))


def test_from_json_obj_deque():
    """ Checks that dequeues are decoded correctly. """
    encoding = [1, 2, 3]
    val = from_json_obj(encoding, Deque[int])
    assert isinstance(val, deque)
    assert len(val) == 3
    assert all(val[i] == encoding[i] for i in range(len(encoding)))


def test_from_json_obj_set():
    """ Checks that dequeues are decoded correctly. """
    encoding = [1, 2, 3]
    val = from_json_obj(encoding, Set[int])
    assert isinstance(val, set)
    assert set(val) == set(encoding)


def test_from_json_obj_frozenset():
    """ Checks that dequeues are decoded correctly. """
    encoding = [1, 2, 3]
    val = from_json_obj(encoding, FrozenSet[int])
    assert isinstance(val, frozenset)
    assert set(val) == set(encoding)


DICT_ENCODINGS = [
    ({str(i): i for i in range(3)}, Dict[str, int], {str(i): i for i in range(3)}),
    ({str(i): i for i in range(3)}, Mapping[str, int], {str(i): i for i in range(3)}),
    ({str(i): i for i in range(3)}, Dict[Literal["0", "1", "2"], int], {str(i): i for i in range(3)}),
    ({str(i): i for i in range(3)}, Mapping[Literal["0", "1", "2"], int], {str(i): i for i in range(3)}),
    ({EnumT.Red: 0, EnumT.Green: 1, EnumT.Blue: 2}, Dict[EnumT, int], {"Red": 0, "Green": 1, "Blue": 2}),
    ({EnumT.Red: 0, EnumT.Green: 1, EnumT.Blue: 2}, Mapping[EnumT, int], {"Red": 0, "Green": 1, "Blue": 2}),
    ({tuple(j for j in range(i)): i for i in range(3)}, Dict[Tuple[int, ...], int], {json.dumps([j for j in range(i)]): i for i in range(3)}),
    ({tuple(j for j in range(i)): i for i in range(3)}, Mapping[Tuple[int, ...], int], {json.dumps([j for j in range(i)]): i for i in range(3)}),
]

def test_from_json_obj_dicts():
    """ Checks that dicts are decoded into dicts with the correct key/value pairs. """
    def _assert_eq_dicts(d1, d2):
        assert isinstance(d1, dict)
        assert isinstance(d2, dict)
        assert len(d1) == len(d2)
        assert set(d1.keys()) == set(d2.keys())
        assert all(d1[key] == d2[key] for key in d1)
    for val, t, encoding in DICT_ENCODINGS:
        _assert_eq_dicts(from_json_obj(encoding, t), val)

ORDERED_DICT_ENCODINGS = [
    (OrderedDict([(str(i), i) for i in range(3)]), typing.OrderedDict[str, int], OrderedDict([(str(i), i) for i in range(3)])),
    (OrderedDict([(str(i), i) for i in range(3)]), typing.OrderedDict[Literal["0", "1", "2"], int], OrderedDict([(str(i), i) for i in range(3)])),
    (OrderedDict([(EnumT.Red, 0), (EnumT.Green, 1), (EnumT.Blue, 2)]), typing.OrderedDict[EnumT, int], OrderedDict([("Red", 0), ("Green", 1), ("Blue", 2)])),
    (OrderedDict([(tuple(j for j in range(i)), i) for i in range(3)]), typing.OrderedDict[Tuple[int, ...], int], OrderedDict([(json.dumps([j for j in range(i)]), i) for i in range(3)])),
]


def test_from_json_obj_ordereddicts():
    """ Checks that dicts are decoded into dicts with the correct key/value pairs. """
    def _assert_eq_ordered_dicts(d1, d2):
        assert isinstance(d1, OrderedDict)
        assert isinstance(d2, OrderedDict)
        assert len(d1) == len(d2)
        assert tuple(d1.keys()) == tuple(d2.keys())
        assert all(d1[key] == d2[key] for key in d1)
    for val, t, encoding in ORDERED_DICT_ENCODINGS:
        _assert_eq_ordered_dicts(from_json_obj(encoding, t), val)



class NamedTupleExampleT(NamedTuple):
    name: str
    value: Union[int, float]
    flag: bool = False

class NamedTupleNastyExampleT(NamedTuple):
    name: str
    value: Dict[str, float]
    flag: bool = False

def test_is_json_encodable_namedtuple():
    """
        Tests the encoding of namedtuples.
    """
    t = NamedTupleExampleT("t", 1)
    t_encoding = OrderedDict([("name", "t"), ("value", 1)])
    assert from_json_obj(t_encoding, NamedTupleExampleT) == t
    nt = NamedTupleNastyExampleT("t", {"x": 1.0, "y": 0.0})
    nt_encoding_list = ["t", {"x": Decimal("1.0"), "y": 0}]
    assert from_json_obj(nt_encoding_list, NamedTupleNastyExampleT) == nt
    nt_encoding_list_long = ["t", {"x": Decimal("1.0"), "y": 0}, False]
    assert from_json_obj(nt_encoding_list_long, NamedTupleNastyExampleT) == nt
    try:
        nt_encoding_list_too_long = ["t", {"x": Decimal("1.0"), "y": 0}, False, 1]
        from_json_obj(nt_encoding_list_too_long, NamedTupleNastyExampleT) == nt
    except TypeError:
        assert True


WRONG_TYPE_ENCODINGS = [
    (bool, 1),
    (int, True),
    (float, "hello"),
    (str, 1.5),
    (type(None), 1),
    (None, 1),
    (Decimal, "hi"),
    (EnumT, 1.2),
    (EnumT, "Yellow"),
    (List[int], "hello"),
    (Tuple[int, ...], tuple([1, 2, 3])),
    (Tuple[int, int, int], tuple([1, 2, 3])),
    (Tuple[int, int, int], [1, 2]),
    (Deque[int], deque([1, 2, 3])),
    (Set[int], set([1, 2, 3])),
    (FrozenSet[int], frozenset([1, 2, 3])),
    (Literal[1,2,3], "1.0"),
    (Union[int, float], False),
    (Dict[str, int], [1, 2, 3]),
    (Mapping[str, int], [1, 2, 3]),
    (typing.OrderedDict[str, int], [1, 2, 3]),
    (Dict[str, int], {1: 1, 2: 2, 3: 3}),
    (Mapping[str, int], {1: 1, 2: 2, 3: 3}),
    (typing.OrderedDict[str, int], OrderedDict({1: 1, 2: 2, 3: 3})),
    (NamedTupleExampleT, ("t", 1)),
    (NamedTupleExampleT, OrderedDict([("name", "t"), ("values", [1, 2, 3])])),
    (NamedTupleExampleT, OrderedDict([("name", "t"), ("value", 1), ("blah", None)])),
    (NamedTupleExampleT, OrderedDict([("name", "t")])),
]


def test_from_json_obj_wrong_type():
    """ Test `to_json_obj` on wrong encodings for JSON base types and `decimal.Decimal`. """
    for t, encoding in WRONG_TYPE_ENCODINGS:
        try:
            from_json_obj(encoding, t)
            assert False, "Should not be decoding %s as %s."%(str(encoding), str(t))
        except TypeError:
            assert True


""" Tests for `typing_json.typechecking.is_keyable` """
# pylint: disable = line-too-long, invalid-name, missing-docstring

# standard imports
from typing import Union, Optional, List, Tuple, Set, FrozenSet, Mapping, Dict, OrderedDict, NamedTuple

# external dependencies
from typing_extensions import Literal

# internal imports
from typing_json.typechecking import is_keyable, KEYABLE_BASE_TYPES, TYPECHECKABLE_BASE_TYPES


NONKEYABLE_BASE_TYPES = tuple(t for t in TYPECHECKABLE_BASE_TYPES if t not in KEYABLE_BASE_TYPES)


def failure_callback(message: str) -> None:
    print(message)


def test_is_keyable_basetypes1():
    """
        Tests that base types which are meant
        to be keyable are actually keyable.
    """
    for t in KEYABLE_BASE_TYPES:
        assert is_keyable(t, failure_callback=failure_callback)
    assert is_keyable(None, failure_callback=failure_callback)


def test_is_keyable_basetypes2():
    """
        Tests that base types which are not meant
        to be keyable are actually not keyable.
    """
    for t in NONKEYABLE_BASE_TYPES:
        assert not is_keyable(t, failure_callback=failure_callback)


def test_is_keyable_enums():
    """
        Tests that enums are keyable.
    """
    from enum import Enum
    class Color(Enum):
        # pylint:disable=all
        Red = 0
        Green = 1
        Blue = 2
    assert is_keyable(Color, failure_callback=failure_callback)


def test_is_keyable_collections1():
    """
        Tests that keyable collections of
        keyable base types are keyable.
    """
    for t in KEYABLE_BASE_TYPES:
        assert is_keyable(FrozenSet[t], failure_callback=failure_callback)
        assert is_keyable(Union[t, Tuple[t]], failure_callback=failure_callback)
        assert is_keyable(Optional[t], failure_callback=failure_callback)
        assert is_keyable(Tuple[t, ...], failure_callback=failure_callback)
        assert is_keyable(Tuple[t], failure_callback=failure_callback)
        assert is_keyable(Tuple[t, int], failure_callback=failure_callback)
        assert is_keyable(Tuple[t, int, str], failure_callback=failure_callback)


def test_is_keyable_collections2():
    """
        Tests that non-keyable collections of
        keyable base types are not keyable.
    """
    for t in TYPECHECKABLE_BASE_TYPES:
        assert not is_keyable(Set[t], failure_callback=failure_callback)
        assert not is_keyable(List[t], failure_callback=failure_callback)
        assert not is_keyable(Dict[str, t], failure_callback=failure_callback)
        assert not is_keyable(Dict[t, int], failure_callback=failure_callback)
        assert not is_keyable(OrderedDict[str, t], failure_callback=failure_callback)
        assert not is_keyable(OrderedDict[t, int], failure_callback=failure_callback)
        assert not is_keyable(Mapping[str, t], failure_callback=failure_callback)
        assert not is_keyable(Mapping[t, int], failure_callback=failure_callback)


def test_is_keyable_collections3():
    """
        Tests that keyable collections of
        non-keyable base types are not keyable.
    """
    for t in NONKEYABLE_BASE_TYPES:
        assert not is_keyable(FrozenSet[t], failure_callback=failure_callback)
        assert not is_keyable(Union[t, Tuple[t]], failure_callback=failure_callback)
        assert not is_keyable(Optional[t], failure_callback=failure_callback)
        assert not is_keyable(Tuple[t, ...], failure_callback=failure_callback)
        assert not is_keyable(Tuple[t], failure_callback=failure_callback)
        assert not is_keyable(Tuple[t, int], failure_callback=failure_callback)
        assert not is_keyable(Tuple[t, int, str], failure_callback=failure_callback)


def test_is_keyable_literal():
    """
        Tests that literals are keyable.
    """
    assert is_keyable(Literal["hi", 1, 2.5])


def test_is_keyable_namedtuple():
    """
        Tests that namedtuples are keyable
        if and only if they have keyable
        field type.
    """
    class HashableNamedTupleExampleT(NamedTuple):
        name: Tuple[str, ...]
        value: Union[int, float]
    assert is_keyable(HashableNamedTupleExampleT, failure_callback=failure_callback)
    class NonHashableNamedTupleExampleT(NamedTuple):
        name: List[str]
        value: Union[int, float]
    assert not is_keyable(NonHashableNamedTupleExampleT, failure_callback=failure_callback)



""" Tests for `typing_json.typechecking.is_typecheckable` """
# pylint: disable = line-too-long, invalid-name, missing-docstring

# standard imports
from typing import Union, Optional, List, Tuple, Set, FrozenSet, Mapping, Dict, OrderedDict, NamedTuple

# external dependencies
from typing_extensions import Literal

# internal imports
from typing_json.typechecking import is_typecheckable, TYPECHECKABLE_BASE_TYPES


def failure_callback(message: str) -> None:
    print(message)


def test_is_typecheckable_basetypes1():
    """
        Tests that base types which are meant
        to be typecheckable are actually typecheckable.
    """
    for t in TYPECHECKABLE_BASE_TYPES:
        assert is_typecheckable(t, failure_callback=failure_callback)
    assert is_typecheckable(None, failure_callback=failure_callback)


def test_is_typecheckable_basetypes2():
    """
        Tests that types which are not meant to be
        typecheckable are not typecheckable.
    """
    class NonTypechekableT:
        name: str
        val: int
    assert not is_typecheckable(NonTypechekableT, failure_callback=failure_callback)


def test_is_typecheckable_enums():
    """
        Tests that enums are typecheckable.
    """
    from enum import Enum
    class Color(Enum):
        # pylint:disable=all
        Red = 0
        Green = 1
        Blue = 2
    assert is_typecheckable(Color, failure_callback=failure_callback)


def test_is_typecheckable_collections1():
    """
        Tests that collections of typecheckable base types
        are typecheckable.
    """
    for t in TYPECHECKABLE_BASE_TYPES:
        assert is_typecheckable(FrozenSet[t], failure_callback=failure_callback)
        assert is_typecheckable(Union[t, Tuple[t]], failure_callback=failure_callback)
        assert is_typecheckable(Optional[t], failure_callback=failure_callback)
        assert is_typecheckable(Tuple[t, ...], failure_callback=failure_callback)
        assert is_typecheckable(Tuple[t], failure_callback=failure_callback)
        assert is_typecheckable(Tuple[t, int], failure_callback=failure_callback)
        assert is_typecheckable(Tuple[t, int, str], failure_callback=failure_callback)
        assert is_typecheckable(Set[t], failure_callback=failure_callback)
        assert is_typecheckable(List[t], failure_callback=failure_callback)
        assert is_typecheckable(Dict[str, t], failure_callback=failure_callback)
        assert is_typecheckable(Dict[t, int], failure_callback=failure_callback)
        assert is_typecheckable(OrderedDict[str, t], failure_callback=failure_callback)
        assert is_typecheckable(OrderedDict[t, int], failure_callback=failure_callback)
        assert is_typecheckable(Mapping[str, t], failure_callback=failure_callback)
        assert is_typecheckable(Mapping[t, int], failure_callback=failure_callback)


def test_is_typecheckable_collections2():
    """
        Tests that collections of non-typecheckable
        types are not typecheckable.
    """
    class NonTypechekableT:
        name: str
        val: int
    assert not is_typecheckable(FrozenSet[NonTypechekableT], failure_callback=failure_callback)
    assert not is_typecheckable(Union[NonTypechekableT, Tuple[int]], failure_callback=failure_callback)
    assert not is_typecheckable(Optional[NonTypechekableT], failure_callback=failure_callback)
    assert not is_typecheckable(Tuple[NonTypechekableT, ...], failure_callback=failure_callback)
    assert not is_typecheckable(Tuple[NonTypechekableT], failure_callback=failure_callback)
    assert not is_typecheckable(Tuple[NonTypechekableT, int], failure_callback=failure_callback)
    assert not is_typecheckable(Set[NonTypechekableT], failure_callback=failure_callback)
    assert not is_typecheckable(List[NonTypechekableT], failure_callback=failure_callback)
    assert not is_typecheckable(Dict[str, NonTypechekableT], failure_callback=failure_callback)
    assert not is_typecheckable(Dict[NonTypechekableT, int], failure_callback=failure_callback)
    assert not is_typecheckable(OrderedDict[str, NonTypechekableT], failure_callback=failure_callback)
    assert not is_typecheckable(OrderedDict[NonTypechekableT, int], failure_callback=failure_callback)
    assert not is_typecheckable(Mapping[str, NonTypechekableT], failure_callback=failure_callback)
    assert not is_typecheckable(Mapping[NonTypechekableT, int], failure_callback=failure_callback)


def test_is_typecheckable_literal():
    """
        Tests that literals are typecheckable.
    """
    assert is_typecheckable(Literal["hi", 1, 2.5])


def test_is_typecheckable_namedtuple():
    """
        Tests that namedtuples are typecheckable if
        and only if their attributes are of typecheckable type.
    """
    class HashableNamedTupleExampleT(NamedTuple):
        name: Tuple[str, ...]
        value: Union[int, float]
    assert is_typecheckable(HashableNamedTupleExampleT, failure_callback=failure_callback)
    class NonHashableNamedTupleExampleT(NamedTuple):
        name: List[str]
        value: Union[int, float]
    assert is_typecheckable(NonHashableNamedTupleExampleT, failure_callback=failure_callback)
    class NonTypechekableT:
        name: str
        val: int
    class NonTypecheckableNamedTupleExampleT(NamedTuple):
        name: List[NonTypechekableT]
        value: Union[int, float]
    assert not is_typecheckable(NonTypecheckableNamedTupleExampleT, failure_callback=failure_callback)



# pylint:disable = line-too-long, invalid-name,
"""
    The `typing_json` library offers type-aware JSON encoding and decoding functionalities,
    on top of the ones offered by the builtin `json` library.

    The functions `typing_json.dump`, `typing_json.dumps`, `typing_json.load` and `typing_json.loads` mirror the
    functionality of their `json` counterparts, adding type-aware encoding/decoding and runtime type-checking of decoded objects.

    The types to be used for type-aware encoding/decoding and runtime type-checking are passed by the additional parameters `encoded_type`
    of `dump`/`dumps` and `decoded_type` of `load`/`loads`.
    Supported types include JSON basic types, `decimal.Decimal`, typed collections from the `typing` module, literal types, union types,
    optional types and (certain) typed namedtuples.
    For a complete list of supported types, see the documentation of `typing_json.encoding.is_json_encodable` from the `typing_json.encoding` module.

    The function `typing_json.typechecking.is_instance` (which can be imported directly as `from typing_json import is_instance`) extends the
    functionality of the builtin `isinstance` to include all the additional types supported by this library.

    The functions `typing_json.load` and `typing_json.loads` deviate from their `json` counterparts in that the default value for the optional
    `parse_float` parameter is `decimal.Decimal`, rather than `float`.
    The function `typing_json.typechecking.is_instance` deviates from the behaviour of the builtin `isinstance` on the literals `True`/`False`,
    which are not deemed of type `int`, and on integral instances of `decimal.Decimal`, which are deemed of type `int` if the optional parameter
    `cast_decimal` is set to `True` (its default value).

    (Version: 0.1.0)
"""

# standard imports
import collections
from decimal import Decimal
import json
from typing import Any, List, Tuple, Type

# internal imports
from typing_json.decoding import from_json_obj
from typing_json.encoding import is_json_encodable, to_json_obj
from typing_json.typechecking import is_instance, is_keyable, is_namedtuple, is_typecheckable


name: str = "typing_json"
__version__: str = "0.1.1.post1"

def dump(obj: Any, encoded_type: Type, fp, skipkeys=False, ensure_ascii=True, check_circular=True, allow_nan=True, cls=None, indent=None, separators=None, default=None, sort_keys=False, **kw) -> None:
    # pylint: disable = too-many-arguments
    """
        Encodes `obj` as a JSON object using `encoded_type` as a type hint, then calls `json.dump`.

        Raises `TypeError` is `encoded_type` is not JSON-encodable according to `typing_json.encoding.is_json_encodable`.
    """
    if not is_json_encodable(encoded_type):
        raise TypeError("Type %s is not json-encodable."%str(encoded_type))
    json_obj = to_json_obj(obj, encoded_type)
    return json.dump(json_obj, fp, skipkeys=skipkeys, ensure_ascii=ensure_ascii, check_circular=check_circular, allow_nan=allow_nan, cls=cls, indent=indent, separators=separators, default=default, sort_keys=sort_keys, **kw)


def dumps(obj: Any, encoded_type: Type, skipkeys=False, ensure_ascii=True, check_circular=True, allow_nan=True, cls=None, indent=None, separators=None, default=None, sort_keys=False, **kw) -> str:
    # pylint: disable = too-many-arguments
    """
        Encodes `obj` as a JSON object using `encoded_type` as a type hint, then calls `json.dumps`.

        Raises `TypeError` is `encoded_type` is not JSON-encodable according to `typing_json.encoding.is_json_encodable`.
    """
    if not is_json_encodable(encoded_type):
        raise TypeError("Type %s is not json-encodable."%str(encoded_type))
    json_obj = to_json_obj(obj, encoded_type)
    return json.dumps(json_obj, skipkeys=skipkeys, ensure_ascii=ensure_ascii, check_circular=check_circular, allow_nan=allow_nan, cls=cls, indent=indent, separators=separators, default=default, sort_keys=sort_keys, **kw)


def load(fp, decoded_type: Type, cast_decimal: bool = True, cls=None, parse_float=Decimal, parse_int=None, parse_constant=None, **kw) -> Any:
    # pylint: disable = too-many-arguments
    """
        Calls `json.load`, then decodes `obj` from the resulting JSON object using `decoded_type` as a type hint.

        Raises `TypeError` is `decoded_type` is not JSON-encodable according to `typing_json.encoding.is_json_encodable`.
    """
    if not is_json_encodable(decoded_type):
        raise TypeError("Type %s is not json-encodable."%str(decoded_type))
    obj = json.load(fp, cls=cls, parse_float=parse_float, parse_int=parse_int, parse_constant=parse_constant, object_pairs_hook=collections.OrderedDict, **kw)
    return from_json_obj(obj, decoded_type, cast_decimal=cast_decimal)


def loads(s: str, decoded_type: Type, cast_decimal: bool = True, encoding=None, cls=None, parse_float=Decimal, parse_int=None, parse_constant=None, **kw) -> Any:
    # pylint: disable = too-many-arguments
    """
        Calls `json.load`, then decodes `obj` from the resulting JSON object using `decoded_type` as a type hint.

        Raises `TypeError` is `decoded_type` is not JSON-encodable according to `typing_json.encoding.is_json_encodable`.
    """
    if not is_json_encodable(decoded_type):
        raise TypeError("Type %s is not json-encodable."%str(decoded_type))
    obj = json.loads(s, encoding=encoding, cls=cls, parse_float=parse_float, parse_int=parse_int, parse_constant=parse_constant, object_pairs_hook=collections.OrderedDict, **kw)
    return from_json_obj(obj, decoded_type, cast_decimal=cast_decimal)



""" Tests for `typing_json.typechecking.is_instance` """
# pylint: disable = line-too-long, invalid-name, missing-docstring

# standard imports
import typing
from typing import Any, Union, Optional, List, Tuple, Set, FrozenSet, Mapping, Dict, NamedTuple, Deque
from collections import deque, OrderedDict
from decimal import Decimal

# external dependencies
from typing_extensions import Literal

# internal imports
from typing_json.typechecking import is_instance, TYPECHECKABLE_BASE_TYPES


def failure_callback(message: str) -> None:
    print(message)


def test_is_instance_basetypes1():
    """
        Checks that values which should be instances of
        basic types are actually recognised as instances.
    """
    values = {
        bool: True,
        int: 1,
        float: 1.0,
        complex: 1.0j,
        str: "hello",
        bytes: b"hello",
        bytearray: bytearray(b"hello"),
        memoryview: memoryview(b"hello"),
        list: ["hello", 0],
        tuple: ("hello", 0),
        range: range(15),
        slice: slice(15),
        set: set(["hello", 0]),
        frozenset: frozenset(["hello", 0]),
        dict: {"str": "hello", "int": 0},
        type: str,
        None: None,
        Decimal: Decimal('1.0'),
    }
    for t, obj in values.items():
        assert is_instance(obj, t, failure_callback=failure_callback)
        assert is_instance(obj, Any, failure_callback=failure_callback)

def test_is_instance_basetypes2():
    """
        Checks that values which should not be instances of
        basic types are actually not recognised as instances.
    """
    values = {
        bool: "False", # bool inherits from int, see https://www.python.org/dev/peps/pep-0285/
        int: "1",
        float: "1.0",
        complex: "1.0j",
        str: 1,
        bytes: "hello",
        bytearray: "hello",
        memoryview: "hello",
        list: ("hello", 0),
        tuple: ["hello", 0],
        range: slice(15),
        slice: range(15),
        set: frozenset(["hello", 0]),
        frozenset: set(["hello", 0]),
        dict: ["str", "hello", "int",  0],
        type: 1,
        None: 1,
        Decimal: 1.0,
    }
    for t, obj in values.items():
        assert not is_instance(obj, t, failure_callback=failure_callback)
    assert is_instance(1, float)


def test_is_instance_bool():
    """
        Tests `is_instance` specifically on `bool` and `int`.
        The issue is that `bool` inherits from `int`,
        see https://www.python.org/dev/peps/pep-0285/
    """
    assert not is_instance(True, int)
    assert not is_instance(False, int)
    assert not is_instance(0, bool)
    assert not is_instance(1, bool)


def test_is_instance_enums():
    """ Tests `is_instance` on enums. """
    from enum import Enum
    class Color(Enum):
        # pylint:disable=all
        Red = 0
        Green = 1
        Blue = 2
    class Shape(Enum):
        # pylint:disable=all
        Circle = 0
        Square = 1
    assert is_instance(Color.Red, Color, failure_callback=failure_callback)
    assert not is_instance(0, Color, failure_callback=failure_callback)
    assert not is_instance(Shape.Circle, Color, failure_callback=failure_callback)


def test_is_instance_namedtuple():
    """ Tests `is_instance` on namedtuples. """
    class NamedTupleExample1T(NamedTuple):
        # pylint:disable=all
        name: str
        value: int
    class NamedTupleExample2T(NamedTuple):
        # pylint:disable=all
        name: str
        value: int
    class ClassExampleT:
        # pylint:disable=all
        name: str
        value: int
        def __init__(self, name, value):
            self.name = name
            self.value = value
    t1 = NamedTupleExample1T("t1", 1)
    t2 = NamedTupleExample2T("t2", 1)
    c = ClassExampleT("c", 1)
    assert is_instance(t1, NamedTupleExample1T, failure_callback=failure_callback)
    assert not is_instance(t2, NamedTupleExample1T, failure_callback=failure_callback)
    assert not is_instance(c, NamedTupleExample1T, failure_callback=failure_callback)
    assert not is_instance(t1._replace(value="1"), NamedTupleExample1T, failure_callback=failure_callback)


def test_is_instance_union():
    """ Tests `is_instance` on unions and optional types. """
    assert is_instance(True, Optional[bool], failure_callback=failure_callback)
    assert is_instance(None, Optional[bool], failure_callback=failure_callback)
    assert not is_instance("hi", Optional[bool], failure_callback=failure_callback)
    assert is_instance(1, Union[int, str], failure_callback=failure_callback)
    assert is_instance("hi", Union[int, str], failure_callback=failure_callback)
    assert not is_instance(range(15), Union[int, str], failure_callback=failure_callback)


def test_is_instance_literal():
    """ Tests `is_instance` on literal types. """
    assert is_instance("hi", Literal["hi", 1, 2.5], failure_callback=failure_callback)
    assert is_instance(1, Literal["hi", 1, 2.5], failure_callback=failure_callback)
    assert is_instance(2.5, Literal["hi", 1, 2.5], failure_callback=failure_callback)
    assert not is_instance("hello", Literal["hi", 1, 2.5], failure_callback=failure_callback)
    assert not is_instance(2, Literal["hi", 1, 2.5], failure_callback=failure_callback)
    assert not is_instance(1.5, Literal["hi", 1, 2.5], failure_callback=failure_callback)


def test_is_instance_list():
    """ Tests `is_instance` on lists. """
    assert is_instance([], List[int], failure_callback=failure_callback)
    assert is_instance([0, 1, 2], List[int], failure_callback=failure_callback)
    assert not is_instance((0, 1, 2), List[int], failure_callback=failure_callback)
    assert not is_instance([0, 1, "2"], List[int], failure_callback=failure_callback)
    assert is_instance([0, 1, "2"], List[Union[int, str]], failure_callback=failure_callback)


def test_is_instance_tuple1():
    """ Tests `is_instance` on variadic tuples. """
    assert is_instance(tuple(), Tuple[int, ...], failure_callback=failure_callback)
    assert is_instance((0, 1, 2), Tuple[int, ...], failure_callback=failure_callback)
    assert not is_instance([0, 1, 2], Tuple[int, ...], failure_callback=failure_callback)
    assert not is_instance((0, 1, "2"), Tuple[int, ...], failure_callback=failure_callback)
    assert is_instance((0, 1, "2"), Tuple[Union[int, str], ...], failure_callback=failure_callback)


def test_is_instance_tuple2():
    """ Tests `is_instance` on fixed-length tuples. """
    assert not is_instance(tuple(), Tuple[int, int, int], failure_callback=failure_callback)
    assert not is_instance((0, 1), Tuple[int, int, int], failure_callback=failure_callback)
    assert is_instance((0, 1, 2), Tuple[int, int, int], failure_callback=failure_callback)
    assert not is_instance((0, 1, 2, 3), Tuple[int, int, int], failure_callback=failure_callback)
    assert not is_instance([0, 1, 2], Tuple[int, int, int], failure_callback=failure_callback)
    assert not is_instance((0, 1, "2"), Tuple[int, int, int], failure_callback=failure_callback)
    assert is_instance((0, 1, "2"), Tuple[int, int, str], failure_callback=failure_callback)


def test_is_instance_set():
    """ Tests `is_instance` on sets. """
    assert is_instance(set(), Set[int], failure_callback=failure_callback)
    assert is_instance({0, 1, 2}, Set[int], failure_callback=failure_callback)
    assert not is_instance([0, 1, 2], Set[int], failure_callback=failure_callback)
    assert not is_instance({0, 1, "2"}, Set[int], failure_callback=failure_callback)
    assert is_instance({0, 1, "2"}, Set[Union[int, str]], failure_callback=failure_callback)


def test_is_instance_frozenset():
    """ Tests `is_instance` on frozensets. """
    assert is_instance(frozenset(), FrozenSet[int], failure_callback=failure_callback)
    assert is_instance(frozenset({0, 1, 2}), FrozenSet[int], failure_callback=failure_callback)
    assert not is_instance({0, 1, 2}, FrozenSet[int], failure_callback=failure_callback)
    assert not is_instance(frozenset({0, 1, "2"}), FrozenSet[int], failure_callback=failure_callback)
    assert is_instance(frozenset({0, 1, "2"}), FrozenSet[Union[int, str]], failure_callback=failure_callback)


def test_is_instance_deque():
    """ Tests `is_instance` on dequeues. """
    assert is_instance(deque(), Deque[int], failure_callback=failure_callback)
    assert is_instance(deque([0, 1, 2]), Deque[int], failure_callback=failure_callback)
    assert not is_instance([0, 1, 2], Deque[int], failure_callback=failure_callback)
    assert not is_instance(deque([0, 1, "2"]), Deque[int], failure_callback=failure_callback)
    assert is_instance(deque([0, 1, "2"]), Deque[Union[int, str]], failure_callback=failure_callback)


def test_is_instance_dict():
    """ Tests `is_instance` on dictionaries. """
    assert is_instance(dict(), Dict[str, int], failure_callback=failure_callback)
    assert is_instance({"0": 0, "1": 1, "2": 2}, Dict[str, int], failure_callback=failure_callback)
    assert not is_instance([0, 1, 2], Dict[str, int], failure_callback=failure_callback)
    assert not is_instance({"0": 0, "1": 1, "2": "2"}, Dict[str, int], failure_callback=failure_callback)
    assert not is_instance({"0": 0, "1": 1, 2: 2}, Dict[str, int], failure_callback=failure_callback)
    assert is_instance({"0": 0, "1": 1, "2": "2"}, Dict[str, Union[str, int]], failure_callback=failure_callback)
    assert is_instance({"0": 0, "1": 1, 2: 2}, Dict[Union[str, int], int], failure_callback=failure_callback)


def test_is_instance_ordereddict():
    """ Tests `is_instance` on ordered dictionaries. """
    assert is_instance(OrderedDict(), typing.OrderedDict[str, int], failure_callback=failure_callback)
    assert is_instance(OrderedDict({"0": 0, "1": 1, "2": 2}), typing.OrderedDict[str, int], failure_callback=failure_callback)
    assert not is_instance([0, 1, 2], typing.OrderedDict[str, int], failure_callback=failure_callback)
    assert not is_instance(OrderedDict({"0": 0, "1": 1, "2": "2"}), typing.OrderedDict[str, int], failure_callback=failure_callback)
    assert not is_instance(OrderedDict({"0": 0, "1": 1, 2: 2}), typing.OrderedDict[str, int], failure_callback=failure_callback)
    assert is_instance(OrderedDict({"0": 0, "1": 1, "2": "2"}), typing.OrderedDict[str, Union[str, int]], failure_callback=failure_callback)
    assert is_instance(OrderedDict({"0": 0, "1": 1, 2: 2}), typing.OrderedDict[Union[str, int], int], failure_callback=failure_callback)


def test_is_instance_mapping():
    """ Tests `is_instance` on read-only dictionaries. """
    # dict
    assert is_instance(dict(), Mapping[str, int], failure_callback=failure_callback)
    assert is_instance({"0": 0, "1": 1, "2": 2}, Mapping[str, int], failure_callback=failure_callback)
    assert not is_instance([0, 1, 2], Mapping[str, int], failure_callback=failure_callback)
    assert not is_instance({"0": 0, "1": 1, "2": "2"}, Mapping[str, int], failure_callback=failure_callback)
    assert not is_instance({"0": 0, "1": 1, 2: 2}, Mapping[str, int], failure_callback=failure_callback)
    assert is_instance({"0": 0, "1": 1, "2": "2"}, Mapping[str, Union[str, int]], failure_callback=failure_callback)
    assert is_instance({"0": 0, "1": 1, 2: 2}, Mapping[Union[str, int], int], failure_callback=failure_callback)
    # OrderedDict
    assert is_instance(OrderedDict(), Mapping[str, int], failure_callback=failure_callback)
    assert is_instance(OrderedDict({"0": 0, "1": 1, "2": 2}), Mapping[str, int], failure_callback=failure_callback)
    assert not is_instance([0, 1, 2], Mapping[str, int], failure_callback=failure_callback)
    assert not is_instance(OrderedDict({"0": 0, "1": 1, "2": "2"}), Mapping[str, int], failure_callback=failure_callback)
    assert not is_instance(OrderedDict({"0": 0, "1": 1, 2: 2}), Mapping[str, int], failure_callback=failure_callback)
    assert is_instance(OrderedDict({"0": 0, "1": 1, "2": "2"}), Mapping[str, Union[str, int]], failure_callback=failure_callback)
    assert is_instance(OrderedDict({"0": 0, "1": 1, 2: 2}), Mapping[Union[str, int], int], failure_callback=failure_callback)


def test_is_instance_unsupported_type():
    class C:
        # pylint: disable = all
        name: str
        value: int
        def __init__(self, name: str, value: int):
            self.name = name
            self.value = value
    c = C("c", 0)
    try:
        is_instance(c, C, failure_callback=failure_callback)
        assert False
    except TypeError:
        assert True

def test_is_instance_decimal1():
    assert is_instance(Decimal("1.0"), int, failure_callback=failure_callback, cast_decimal=True)
    assert is_instance(Decimal(1.0), int, failure_callback=failure_callback, cast_decimal=True)
    assert is_instance(Decimal("1"), int, failure_callback=failure_callback, cast_decimal=True)
    assert is_instance(Decimal(1), int, failure_callback=failure_callback, cast_decimal=True)

def test_is_instance_decimal2():
    assert not is_instance(Decimal("1.0"), int, failure_callback=failure_callback, cast_decimal=False)
    assert not is_instance(Decimal(1.0), int, failure_callback=failure_callback, cast_decimal=False)
    assert not is_instance(Decimal("1"), int, failure_callback=failure_callback, cast_decimal=False)
    assert not is_instance(Decimal(1), int, failure_callback=failure_callback, cast_decimal=False)

def test_is_instance_decimal3():
    assert not is_instance(Decimal("1.1"), int, failure_callback=failure_callback, cast_decimal=True)
    assert not is_instance(Decimal(1.000000000000001), int, failure_callback=failure_callback, cast_decimal=True)
    assert is_instance(Decimal("1.1"), float, failure_callback=failure_callback, cast_decimal=True)
    assert is_instance(Decimal(1.000000000000001), float, failure_callback=failure_callback, cast_decimal=True)
    assert is_instance(Decimal("1.0"), float, failure_callback=failure_callback, cast_decimal=True)
    assert is_instance(Decimal(1), float, failure_callback=failure_callback, cast_decimal=True)

def test_is_instance_decimal5():
    assert not is_instance(Decimal("1.1"), int, failure_callback=failure_callback, cast_decimal=False)
    assert not is_instance(Decimal(1.000000000000001), int, failure_callback=failure_callback, cast_decimal=False)
    assert not is_instance(Decimal("1.1"), float, failure_callback=failure_callback, cast_decimal=False)
    assert not is_instance(Decimal(1.000000000000001), float, failure_callback=failure_callback, cast_decimal=False)
    assert not is_instance(Decimal("1.0"), float, failure_callback=failure_callback, cast_decimal=False)
    assert not is_instance(Decimal(1), float, failure_callback=failure_callback, cast_decimal=False)



#pylint:disable = line-too-long, invalid-name
"""
    The `typing_json.encoding` provides functionality for type-aware JSON-encoding of objects.

    The core functionality is provided by `typing_json.encoding.to_json_obj`, which JSON-encodes instances
    of basic JSON types, typed collections from the `typing` module, literal types, union types, optional types
    and (certain) typed namedtuples.
    The JSON-encoding preserves all information necessary to reconstruct the at decoding time (cf. `typing_json.decoding.from_json_obj`).

    (Version: 0.1.0)
"""

# standard imports
from collections import deque, OrderedDict
from collections.abc import Mapping
from decimal import Decimal
from enum import EnumMeta
import json
from typing import Any, Callable, List, Optional, Union, Type

# external dependencies
from typing_extensions import Literal

# internal imports
from typing_json.typechecking import is_instance, is_keyable, is_namedtuple, is_typecheckable, JSON_BASE_TYPES, short_str


_UNREACHABLE_ERROR_MSG = "Should never reach this point, please open an issue on GitHub."


def _not_json_encodable(message: str, failure_callback: Optional[Callable[[str], None]]) -> Literal[False]:
    """ Utility message to fail (return `False`) by first calling an optional failure callback. """
    if failure_callback:
        failure_callback(message)
    return False


def is_json_encodable(t: Type, failure_callback: Optional[Callable[[str], None]] = None) -> bool:
    """
        Checks whether a type `t` can be encoded into JSON (or decoded from JSON) using the `typing_json` library.

        The optional parameter `failure_callback` can be used to collect a detailed trace of
        the reasons behind this method returning `False` on a given type `t`.

        Currently, a type `t` is JSON encodable according to this method if it is typecheckable according to
        `typing_json.typechecking.is_typecheckable` and it satisfies one of the following conditions:

        - if `t` is one of the JSON basic types `bool`, `int`, `float`, `str`, `NoneType`;
        - if `t` is a `decimal.Decimal`;
        - if `t` is `None` (used as an alias for `NoneType`);
        - if `t` is an enum (i.e. `isinstance(t, EnumMeta)`);
        - if `t` is a namedtuple according to `typing_json.typechecking.is_namedtuple` and all its fields are JSON encodable;
        - if `t` is one of `typing.List`, `typing.Set`, `typing.FrozenSet`, `typing.Deque`, `typing.Optional` or a variadic `typing.Tuple` and its generic type argument is JSON encodable;
        - if `t` is a `typing.Union` or a fixed-length `typing.Tuple` and all of its generic type arguments are JSON encodable;
        - if `t` is a `typing.Dict`, `typing.OrderedDict` or `typing.Mapping`, its generic key type is keyable (according to `typing_json.typechecking.is_keyable`) and both its generic key and value types are JSON encodable;
        - if `t` is a `typing_extensions.Literal` and all of its literal arguments are of JSON basic type.

    """
    # pylint: disable = too-many-return-statements, too-many-branches
    if not is_typecheckable(t, failure_callback=failure_callback):
        # only typecheckable types are encodable
        return _not_json_encodable("Type %s is not typecheckable."%str(t), failure_callback=failure_callback)
    if t in JSON_BASE_TYPES:
        # JSON basic types are encodable
        return True
    if t is Decimal:
        # `decimal.Decimal` is encodable
        return True
    if t is None:
        # `None` canbe used as an alias for class `NoneType`
        return True
    if isinstance(t, EnumMeta):
        # enums are encodable
        return True
    if is_namedtuple(t):
        field_types = getattr(t, "_field_types")
        if all(is_json_encodable(field_types[field], failure_callback=failure_callback) for field in field_types):
            # namedtuples are encodable if all their fields are of encodable types
            return True
        return _not_json_encodable("Not all fields of namedtuple %s are json-encodable."%str(t), failure_callback=failure_callback)
    if hasattr(t, "__origin__") and hasattr(t, "__args__"):
        # `typing` generics
        if t.__origin__ in (list, set, frozenset, deque, Optional):
            if is_json_encodable(t.__args__[0], failure_callback=failure_callback):
                # `typing.List`, `typing.Set`, `typing.FrozenSet`, `typing.Deque` and `typing.Optional` are encodable if their generic type argument is encodable
                return True
            return _not_json_encodable("Type of elements in %s is not json-encodable."%str(t), failure_callback=failure_callback)
        if t.__origin__ is tuple:
            # `typing.Tuple`
            if len(t.__args__) == 2 and t.__args__[1] is ...: # pylint:disable=no-else-return
                if is_json_encodable(t.__args__[0], failure_callback=failure_callback):
                    # variadic `typing.Tuple` are encodable if their generic type argument is encodable
                    return True
                return _not_json_encodable("Type of elements in %s is not json-encodable."%str(t), failure_callback=failure_callback)
            else:
                if all(is_json_encodable(s, failure_callback=failure_callback) for s in t.__args__):
                    # fixed-length `typing.Tuple` are encodable if all their generic type arguments are encodable
                    return True
                return _not_json_encodable("Type of some element in %s is not json-encodable."%str(t), failure_callback=failure_callback)
        if t.__origin__ is Union:
            if all(is_json_encodable(s, failure_callback=failure_callback) for s in t.__args__):
                # `typing.Union` are encodable if all their generic type arguments are encodable
                return True
            return _not_json_encodable("Some type in %s is not json-encodable."%str(t), failure_callback=failure_callback)
        if t.__origin__ in (dict, OrderedDict, Mapping):
            # `typing.Dict`, `typing.OrderedDict` and `typing.Mapping` are encodable if their generic key and value types are encodable and their key type is keyable
            if not is_keyable(t.__args__[0], failure_callback=failure_callback):
                return _not_json_encodable("Type of keys in %s is not keyable."%str(t), failure_callback=failure_callback)
            if not is_json_encodable(t.__args__[0], failure_callback=failure_callback):
                return _not_json_encodable("Type of keys in %s is not json-encodable."%str(t), failure_callback=failure_callback)
            if not is_json_encodable(t.__args__[1], failure_callback=failure_callback):
                return _not_json_encodable("Type of values in %s is not json-encodable."%str(t), failure_callback=failure_callback)
            return True
        if t.__origin__ is Literal:
            # `typing_extensions.Literal` are encodable as long as their literals are JSON basic types, which is always the case if they are typecheckable.
            return True
    return False


def to_json_obj(obj: Any, t: Type, use_decimal: bool = False) -> Any:
    """
        Encodes an instance `obj` of typecheckable type `t` into a JSON object.
        The optional `use_decimal` parameter can be used to specify that instances of
        `decimal.Decimal` can be used in the output: if `False`, they are converted to strings.
        This method raises `TypeError` if type `t` is not typecheckable according to `typing_json.typechecking.is_typecheckable`.
        This method raises `TypeError` if `obj` is not of type `t` according to `typing_json.typechecking.is_instance`.

        The optional parameter `failure_callback` can be used to collect a detailed trace of
        the reasons behind this method returning `False` on a given type `t`.

        Currently, this method acts as follows on an instance `obj` of type `t`:

        - if `t` is one of the JSON basic types `bool`, `int`, `float`, `str`, `NoneType`, the instance `obj` is returned unchanged;
        - if `t` is `decimal.Decimal` and `use_decimal` is `False` (default), `str(obj)` is returned;
        - if `t` is `decimal.Decimal` and `use_decimal` is `True`, `obj` is returned unchanged;
        - if `t` is `None` (used as an alias for `NoneType`), `None` is returned;
        - if `t` is an enum (i.e. `isinstance(t, EnumMeta)`), the enum value name `obj._name_` is returned;
        - if `t` is a namedtuple according to `typing_json.typechecking.is_namedtuple` and all its fields are JSON encodable, this method is called recursively on all field values and then an ordered dictionary is returned with the field names as names and the JSON-encoded field values as corresponding values;
        - if `t` is `typing.Union`, the generic type arguments in the union are tried one after the other until a `u` is found such that `is_instance(obj, u)`, then `obj` is JSON-encoded using `u` as its type.
        - if `t` is a `typing_extensions.Literal`, `obj` is returned unchanged;
        - if `t` is one of `typing.List`, `typing.Set`, `typing.FrozenSet`, `typing.Deque` or `typing.Tuple`, a list is returned containing the elements of the original collection, recursively JSON-encoded;
        - if `t` is a `typing.Dict` or `typing.Mapping`, a dictionary (`dict`) is returned with JSON-encoded values from the original dictionary/mapping, associated to either then JSON-encoded keys or a stringified version of the JSON-encoded keys (cf. below);
        - if `t` is `typing.OrderedDict`, an ordered dictionary (`collections.OrderedDict`) is returned with JSON-encoded values from the original dictionary/mapping, associated to either then JSON-encoded keys or a stringified version of the JSON-encoded keys (cf. below).

        In the case of dictionaries, it is not necessarily the case keys will be compatible with the JSON specification in their JSON-encoded form.
        When encoding dictionaries, the keys used in the encoding follow the following criteria:

        - if the key type is a JOSN basic type, `decimal.Decimal` or an enumeration type, the JSON encoding of the keys is used;
        - otherwise, the stringified version of the JSON encoding (using `json.dumps`) is used;

        Literals can only be of JSON basic type.

    """
    # pylint:disable=invalid-name,too-many-return-statements,too-many-branches
    trace: List[str] = []
    def failure_callback(message: str) -> None:
        trace.append(message)
    if not is_json_encodable(t, failure_callback=failure_callback):
        # Argument `t` must be JSON encodable.
        raise TypeError("Type %s is not json-encodable. Trace:\n%s"%(str(t), "\n".join(trace)))
    trace = []
    if not is_instance(obj, t, failure_callback=failure_callback):
        # Argument `obj` must be an instance of argument `t`.
        raise TypeError("Object %s is not of type %s. Trace:\n%s"%(short_str(obj), str(t), "\n".join(trace)))
    if t in JSON_BASE_TYPES:
        # JSON basic types are returned unchanged.
        return obj
    if t is Decimal:
        # If `use_decimal` is `True`, `obj` is returned unchanged:
        if use_decimal:
            return obj
        # If `use_decimal` is `False` (default), instances of `decimal.Decimal` are encoded as strings.
        return str(obj)
    if t in (None, type(None)):
        # `None` can be used as an alias for `NoneType`.
        return None
    if isinstance(t, EnumMeta):
        # Enum values are encoded by their name.
        return obj._name_ # pylint:disable=protected-access
    if is_namedtuple(t):
        # Namedtuples are encoded as ordered dictionaries, with their fields as keys and the JSON-encoded field values as corresponding values.
        field_types = getattr(t, "_field_types")
        json_dict = OrderedDict() # type:ignore
        for field in field_types:
            json_dict[field] = to_json_obj(getattr(obj, field), field_types[field])
        return json_dict
    if hasattr(t, "__origin__") and hasattr(t, "__args__"):
        # Generics from the `typing` module.
        if t.__origin__ is Union:
            # values in a `typing.Union` are JSON-encoded using the first type in the union that the object is found to be an instance of.
            for s in t.__args__:
                if is_instance(obj, s):
                    return to_json_obj(obj, s)
            raise AssertionError(_UNREACHABLE_ERROR_MSG) # pragma: no cover
        if t.__origin__ is Literal:
            # `typing_extensions.Literal` are returned unchanged
            return obj
        if t.__origin__ in (list, set, frozenset, deque):
            # `typing.List`, `typing.Set`, `typing.FrozenSet` and `typing.Deque` are turned into lists, with their elements recursively JSON-encoded
            return [to_json_obj(x, t.__args__[0]) for x in obj]
        if t.__origin__ is tuple:
            # `typing.Tuple` are turned into lists, with their elements recursively JSON-encoded
            if len(t.__args__) == 2 and t.__args__[1] is ...: # pylint:disable=no-else-return
                return [to_json_obj(x, t.__args__[0]) for x in obj]
            else:
                return [to_json_obj(x, t.__args__[i]) for i, x in enumerate(obj)]
        if t.__origin__ in (dict, OrderedDict, Mapping):
            # `typing.Dict` and `typing.Mapping` are turned into dictionaries and `typing.OrderedDict` are turned into ordered dictionaries.
            # The values are recursively JSON-encoded. Keys require special handling.
            fields = [field for field in obj] # pylint: disable = unnecessary-comprehension
            if t.__args__[0] in JSON_BASE_TYPES+(Decimal, None,):
                # Keys of JSON basic types, `decimal.Decimal` and `None` are recursively JSON-encoded.
                # encoded_fields = [field for field in fields] # pylint: disable = unnecessary-comprehension
                encoded_fields = [to_json_obj(field, t.__args__[0]) for field in fields]
            elif (hasattr(t.__args__[0], "__origin__") and t.__args__[0].__origin__ is Literal):
                # Keys of `typing_extensions.Literal` types are recursively JSON-encoded.
                # encoded_fields = [field for field in fields] # pylint: disable = unnecessary-comprehension
                encoded_fields = [to_json_obj(field, t.__args__[0]) for field in fields]
            elif isinstance(t.__args__[0], EnumMeta):
                # Keys of enumeration types are recursively JSON-encoded.
                encoded_fields = [to_json_obj(field, t.__args__[0]) for field in fields]
            else:
                # Keys of any other type are recursively JSON-encoded and then JSON dumped to strings.
                encoded_fields = [json.dumps(to_json_obj(field, t.__args__[0])) for field in fields]
            if t.__origin__ in (dict, Mapping):
                # A `dict`is used for `typing.Dict` and `typing.Mapping`.
                return {encoded_fields[i]: to_json_obj(obj[field], t.__args__[1]) for i, field in enumerate(fields)}
            if t.__origin__ is OrderedDict:
                # A `collections.OrderedDict` is used for `typing.OrderedDict`.
                new_ordered_dict = OrderedDict() # type:ignore
                for i, field in enumerate(fields):
                    new_ordered_dict[encoded_fields[i]] = to_json_obj(obj[field], t.__args__[1])
                return new_ordered_dict
    raise AssertionError(_UNREACHABLE_ERROR_MSG) # pragma: no cover



""" Tests for `typing_json.typechecking.to_json_obj` """
# pylint: disable = line-too-long, invalid-name, missing-docstring

# standard imports
import json
import typing
from typing import Union, Optional, List, Tuple, Set, FrozenSet, Mapping, Dict, NamedTuple, Deque
from decimal import Decimal
from collections import deque, OrderedDict

# external dependencies
from typing_extensions import Literal

# internal imports
from typing_json.typechecking import TYPECHECKABLE_BASE_TYPES
from typing_json.encoding import to_json_obj, JSON_BASE_TYPES

# (bool, int, float, str, type(None))
NONENCODABLE_BASE_TYPES = tuple(t for t in TYPECHECKABLE_BASE_TYPES if t not in JSON_BASE_TYPES+(Decimal,))

def failure_callback(message: str) -> None:
    print(message)


def test_to_json_obj_typeerrors():
    """ Checks that `to_json_obj` raises `TypeError` on non-typecheckable types and when `obj` is not an instance of `t`.
    """
    class NonTypechekableT:
        name: str
        val: int
        def __init__(self, name, val):
            self.name = name
            self.val = val
    x = NonTypechekableT("x", 0)
    try:
        to_json_obj(x, NonTypechekableT)
        assert False
    except TypeError:
        assert True
    try:
        to_json_obj(0, str)
        assert False
    except TypeError:
        assert True


BASETYPES_ENCODINGS = [
    (True, bool, True),
    (1, int, 1),
    (1.5, float, 1.5),
    ("hello", str, "hello"),
    (None, type(None), None),
    (None, None, None),
    (Decimal("1.0"), Decimal, "1.0"),
    (Decimal("1.5"), Decimal, "1.5"),
    (True, Literal[True, 1.5, None], True),
    (1.5, Literal[True, 1.5, None], 1.5),
    (None, Literal[True, 1.5, None], None),
    (True, Union[bool, float, Decimal], True),
    (1.5, Union[bool, float, Decimal], 1.5),
    (Decimal("1.0"), Union[bool, str, Decimal], "1.0"),
    (Decimal("1.5"), Union[bool, str, Decimal], "1.5"),
]


def test_to_json_obj_basetypes():
    """ Test `to_json_obj` on JSON base types and `decimal.Decimal`. """
    for val, t, encoding in BASETYPES_ENCODINGS:
        assert to_json_obj(val, t) == encoding
    assert to_json_obj(Decimal("1.5"), Decimal, use_decimal=True) == Decimal("1.5")


from enum import Enum
class EnumT(Enum):
    # pylint:disable=all
    Red = 0
    Green = 1
    Blue = 2

def test_to_json_obj_enums():
    assert to_json_obj(EnumT.Red, EnumT) == "Red"


SEQUENCES_ENCODINGS = [
    ([0, 1, 2], List[int], [0, 1, 2]),
    ((0, 1, 2), Tuple[int, ...], [0, 1, 2]),
    ((0, 1, 2), Tuple[int, int, int], [0, 1, 2]),
    (deque([0, 1, 2]), Deque[int], [0, 1, 2]),
]


def test_to_json_obj_sequences():
    """ Checks that sequences are encoded into lists with the correct elements. """
    def _assert_eq_lists(l1, l2):
        assert isinstance(l1, list)
        assert isinstance(l2, list)
        assert len(l1) == len(l2)
        assert all(l1[i] == l2[i] for i in range(len(l1)))
        assert all(type(l1[i]) == type(l2[i]) for i in range(len(l1)))
    for val, t, encoding in SEQUENCES_ENCODINGS:
        _assert_eq_lists(to_json_obj(val, t), encoding)


SETS_ENCODINGS = [
    ({0, 1, 2}, Set[int], [0, 1, 2]),
    (frozenset({0, 1, 2}), FrozenSet[int], [0, 1, 2]),
]


def test_to_json_obj_sets():
    """ Checks that sets are encoded into lists with the correct elements. """
    def _assert_eq_sets(l1, l2):
        assert isinstance(l1, list)
        assert isinstance(l2, list)
        assert len(l1) == len(l2)
        assert set(l1) == set(l2)
    for val, t, encoding in SETS_ENCODINGS:
        _assert_eq_sets(to_json_obj(val, t), encoding)


DICT_ENCODINGS = [
    ({str(i): i for i in range(3)}, Dict[str, int], {str(i): i for i in range(3)}),
    ({str(i): i for i in range(3)}, Mapping[str, int], {str(i): i for i in range(3)}),
    ({str(i): i for i in range(3)}, Dict[Literal["0", "1", "2"], int], {str(i): i for i in range(3)}),
    ({str(i): i for i in range(3)}, Mapping[Literal["0", "1", "2"], int], {str(i): i for i in range(3)}),
    ({EnumT.Red: 0, EnumT.Green: 1, EnumT.Blue: 2}, Dict[EnumT, int], {"Red": 0, "Green": 1, "Blue": 2}),
    ({EnumT.Red: 0, EnumT.Green: 1, EnumT.Blue: 2}, Mapping[EnumT, int], {"Red": 0, "Green": 1, "Blue": 2}),
    ({tuple(j for j in range(i)): i for i in range(3)}, Dict[Tuple[int, ...], int], {json.dumps([j for j in range(i)]): i for i in range(3)}),
    ({tuple(j for j in range(i)): i for i in range(3)}, Mapping[Tuple[int, ...], int], {json.dumps([j for j in range(i)]): i for i in range(3)}),
]


def test_to_json_obj_dicts():
    """ Checks that dicts/mappings are encoded into dicts with the correct key/value pairs. """
    def _assert_eq_dicts(d1, d2):
        assert isinstance(d1, dict)
        assert isinstance(d2, dict)
        assert len(d1) == len(d2)
        assert set(d1.keys()) == set(d2.keys())
        assert all(d1[key] == d2[key] for key in d1)
    for val, t, encoding in DICT_ENCODINGS:
        _assert_eq_dicts(to_json_obj(val, t), encoding)


ORDERED_DICT_ENCODINGS = [
    (OrderedDict([(str(i), i) for i in range(3)]), typing.OrderedDict[str, int], OrderedDict([(str(i), i) for i in range(3)])),
    (OrderedDict([(str(i), i) for i in range(3)]), typing.OrderedDict[Literal["0", "1", "2"], int], OrderedDict([(str(i), i) for i in range(3)])),
    (OrderedDict([(EnumT.Red, 0), (EnumT.Green, 1), (EnumT.Blue, 2)]), typing.OrderedDict[EnumT, int], OrderedDict([("Red", 0), ("Green", 1), ("Blue", 2)])),
    (OrderedDict([(tuple(j for j in range(i)), i) for i in range(3)]), typing.OrderedDict[Tuple[int, ...], int], OrderedDict([(json.dumps([j for j in range(i)]), i) for i in range(3)])),
]


def test_to_json_obj_ordereddicts():
    """ Checks that dicts/mappings are encoded into dicts with the correct key/value pairs. """
    def _assert_eq_ordered_dicts(d1, d2):
        assert isinstance(d1, OrderedDict)
        assert isinstance(d2, OrderedDict)
        assert len(d1) == len(d2)
        assert tuple(d1.keys()) == tuple(d2.keys())
        assert all(d1[key] == d2[key] for key in d1)
    for val, t, encoding in ORDERED_DICT_ENCODINGS:
        _assert_eq_ordered_dicts(to_json_obj(val, t), encoding)


def test_is_json_encodable_namedtuple():
    """
        Tests the encoding of namedtuples.
    """
    def _assert_eq_ordered_dicts(d1, d2):
        assert isinstance(d1, OrderedDict)
        assert isinstance(d2, OrderedDict)
        assert len(d1) == len(d2)
        assert tuple(d1.keys()) == tuple(d2.keys())
        assert all(d1[key] == d2[key] for key in d1)
    class NamedTupleExampleT(NamedTuple):
        name: str
        value: Union[int, float]
    t = NamedTupleExampleT("t", 1)
    t_encoding = OrderedDict([("name", "t"), ("value", 1)])
    _assert_eq_ordered_dicts(to_json_obj(t, NamedTupleExampleT), t_encoding)



""" Tests for `typing_json.typechecking.is_namedtuple` """
# pylint: disable = line-too-long, invalid-name, missing-docstring

# standard imports
import sys
import typing
from typing import Any, Union, Optional, List, Tuple, Set, FrozenSet, Mapping, Dict, NamedTuple, Deque
from collections import namedtuple, deque, OrderedDict
from decimal import Decimal

# external dependencies
from typing_extensions import Literal

# internal imports
from typing_json.typechecking import is_namedtuple, is_instance, TYPECHECKABLE_BASE_TYPES


def failure_callback(message: str) -> None:
    print(message)

def test_is_namedtuple_basic():
    """
        Tests that `typing.NamedTuple` are identified correctly,
        while classes, `tuple` and `namedtuple` are not.
    """
    class A(NamedTuple):
        # pylint:disable=all
        name: List[str]
        value: Union[int, float]
    assert is_namedtuple(A, failure_callback=failure_callback)
    a = A(["hi"], 1.1)
    assert is_instance(a, A, failure_callback=failure_callback)
    a = A(0, 1.1) # type:ignore
    assert not is_instance(a, A, failure_callback=failure_callback)
    a = A(["hi"], 1.1)
    class B:
        # pylint:disable=all
        name: List[str]
        def __init__(self, name: List[str]):
            self.name = name
    class Btuple(tuple):
        # pylint:disable=all
        name: List[str]
        def __init__(self, name: List[str]):
            self.name = name
    assert not is_namedtuple(0, failure_callback=failure_callback)
    assert not is_namedtuple(B, failure_callback=failure_callback)
    assert not is_namedtuple(Btuple, failure_callback=failure_callback)
    b = B(["hi"])
    assert not is_instance(b, A, failure_callback=failure_callback)
    C = namedtuple("C", ["name", "value"])
    assert not is_namedtuple(C, failure_callback=failure_callback)

def test_is_namedtuple_fake1():
    """
        Tries to make a fake namedtuple by adding `_field_types`
        to a `tuple`.

        Note: for some odd reason the following test fails in 3.7.1,
        so this test is only executed in Python 3.7.4+.
    """
    C = namedtuple("C", ["name", "value"])
    version_info = sys.version_info
    if version_info[0:3]>= (3, 7, 4):
        C._field_types = {"name": str, "value": int}
        assert is_namedtuple(C, failure_callback=failure_callback)

def test_is_namedtuple_fake2():
    """
        Tries to make a fake `typing.NamedTuple` by adding `_fields`
        to a `tuple`.
    """
    class D1(tuple):
        _fields = []
    assert not is_namedtuple(D1, failure_callback=failure_callback)
    class D2(tuple):
        _fields = ("name", 0)
    assert not is_namedtuple(D2, failure_callback=failure_callback)

def test_is_namedtuple_fake3():
    """
        Tries to make a fake `typing.NamedTuple` by adding
        `_fields` and `_field_types` to a `tuple`,
        but getting things slightly wrong.
    """
    class B:
        # pylint:disable=all
        name: List[str]
        def __init__(self, name: List[str]):
            self.name = name
    class D3(tuple):
        _fields = ("name", "value")
        _field_types = []
    assert not is_namedtuple(D3, failure_callback=failure_callback)
    class D4(tuple):
        _fields = ("name", "value")
        _field_types = {"name": str, "val": int}
    assert not is_namedtuple(D4, failure_callback=failure_callback)
    class D5(tuple):
        _fields = ("name", "value")
        _field_types = {"name": str, "value": int, "val": int}
    assert not is_namedtuple(D5, failure_callback=failure_callback)
    class D6(tuple):
        _fields = ("name", "value")
        _field_types = {"name": str, "value": B}
    assert not is_namedtuple(D6, failure_callback=failure_callback)
    class D6prime(tuple):
        _fields = ("name", "value")
        _field_types = {"name": str, "value": int}
    assert not is_namedtuple(D6prime, failure_callback=failure_callback)

def test_is_namedtuple_fake4():
    """
        Tries to make a fake `typing.NamedTuple` by adding
        `_fields`, `_field_types` and `_field_defaults` to a `tuple`,
        but getting things slightly wrong.
    """
    class D7(tuple):
        _fields = ("name", "value")
        _field_types = {"name": str, "value": int}
        _field_defaults = []
    assert not is_namedtuple(D7, failure_callback=failure_callback)
    class D8(tuple):
        _fields = ("name", "value")
        _field_types = {"name": str, "value": int}
        _field_defaults = {"val": 0}
    assert not is_namedtuple(D8, failure_callback=failure_callback)
    class D9(tuple):
        _fields = ("name", "value")
        _field_types = {"name": str, "value": int}
        _field_defaults = {"value": "bye"}
    assert not is_namedtuple(D9, failure_callback=failure_callback)

def test_is_namedtuple_fake5():
    """
        Tries to make a fake `typing.NamedTuple` by adding
        `_fields`, `_field_types` and `_field_defaults` to a `tuple`,
        but forgets to add the actual fields to the class.
    """
    class A(NamedTuple):
        # pylint:disable=all
        name: str = "a"
        value: int = 0
    assert is_namedtuple(A, failure_callback=failure_callback)
    class FakeA(tuple):
        _fields = ("name", "value")
        _field_types = {"name": str, "value": int}
        _field_defaults = {"name": "a", "value": 0}
    assert not is_namedtuple(FakeA, failure_callback=failure_callback)

def test_is_namedtuple_fake6():
    """
        Tries to make a fake `typing.NamedTuple` by adding
        `_fields`, `_field_types` and `_field_defaults` to a `tuple`
        and adding the actual fields to the class.
        Unfortunately, the fields don't get added to `dir` for a `tuple`.
    """
    class A(NamedTuple):
        # pylint:disable=all
        name: str = "a"
        value: int = 0
    assert is_namedtuple(A, failure_callback=failure_callback)
    class FakeA(tuple):
        name: str
        value: int
        _fields = ("name", "value")
        _field_types = {"name": str, "value": int}
        _field_defaults = {"name": "a", "value": 0}
    assert not is_namedtuple(FakeA, failure_callback=failure_callback)



#pylint:disable = line-too-long, invalid-name
"""
    The `typing_json.decoding` provides functionality for type-aware JSON-decoding of objects.

    The core functionality is provided by `typing_json.decoding.from_json_obj`, which can be used to
    turn JSON objects obtained from `typing_json.encoding.to_json_obj` back into instances of the original types.
    Types supported include JSON basic types, typed collections from the `typing` module, literal types, union types,
    optional types and (certain) typed namedtuples.

    (Version: 0.1.0)
"""

# standard imports
from collections import deque, OrderedDict
from collections.abc import Mapping
from decimal import Decimal, InvalidOperation
from enum import EnumMeta
import json
from typing import Any, List, Union, Type

# external dependencies
from typing_extensions import Literal

# internal imports
from typing_json.typechecking import is_instance, is_namedtuple, JSON_BASE_TYPES, short_str
from typing_json.encoding import is_json_encodable


_UNREACHABLE_ERROR_MSG = "Should never reach this point, please open an issue on GitHub."


def from_json_obj(obj: Any, t: Type, cast_decimal: bool = True) -> Any:
    """
        Decodes a JSON object `obj` into an instance of a typecheckable type `t`.
        This method raises `TypeError` if type `t` is not JSON encodable according to `typing_json.encoding.is_json_encodable`.
        This method also raises `TypeError` if `obj` is not a valid JSON encoding for an instance of type `t`.

        Currently, this method acts as follows on an JSON object `obj` and a JSON-encodable type `t`:

        - if `t` is one of the JSON basic types `bool`, `float`, `str`, `NoneType`, `obj` must be an instance of the type an is returned unchanged;
        - if `t` is of the JSON basic type `int` and the `cast_decimal` parameter is set to `False`, `obj` must be an instance of `int` and is returned unchanged;
        - it `t` is of the JSON basic type `int` and the `cast_decimal` parameter is set to `True` (its default value), `obj` can be either an instance of `int`, in which case it is returned unchanged, or an instance of `decimal.Decimal` encoding an integer, in which case `int(obj)` is returned;
        - if `t` is of the JSON basic type `float` and the `cast_decimal` parameter is set to `False`, `obj` must be an instance of `int` or `float`, and `float(obj)` is returned;
        - it `t` is of the JSON basic type `float` and the `cast_decimal` parameter is set to `True` (its default value), `obj` can be either an instance of `int`, `float` or `decimal.Decimal`, in which case `float(obj)` is returned;
        - if `t` is `None`, used as an alias for `NoneType`, `obj` must be `None` and is returned unchanged;
        - if `t` is `decimal.Decimal` and the `cast_decimal` parameter is set to `False`, `obj` must be either a `decimal.Decimal`, an `int` or a `str` encoding a valid decimal, in which case `decimal.Decimal(obj)` is returned;
        v- if `t` is `decimal.Decimal` and the `cast_decimal` parameter is set to `True`, `obj` must be either a `decimal.Decimal`, an `int`, a `float` or a `str` encoding a valid decimal, in which case `decimal.Decimal(obj)` is returned;
        - if `t` is an enumeration, `obj` must be a key in the dictionary `t.__members__` of names for the enumeration constants, in which case `t.__members__[obj]` is returned;
        - if `t` is a namedtuple (according to `typing_json.typechecking.is_namedtuple`), see below;
        - if `t` is `typing.Union` or `typing.Optional`, try to decoded `obj` using the generic type arguments one after the other, until a suitable one is found;
        - if `t` is `typing_extensions.Literal`, check that `obj` is one of the literals and return it unaltered;
        - if `t` is `typing.List`, check that `obj` is a list and return a list with recursively JSON-decoded elements of `obj` in it;
        - if `t` is `typing.Tuple`, check that `obj` is a list and return a tuple with recursively JSON-decoded elements of `obj` in it;
        - if `t` is `typing.Deque`, check that `obj` is a list and return a deque with recursively JSON-decoded elements of `obj` in it;
        - if `t` is `typing.Set`, check that `obj` is a list and return a set with recursively JSON-decoded elements of `obj` in it;
        - if `t` is `typing.FrozenSet`, check that `obj` is a list and return a frozenset with recursively JSON-decoded elements of `obj` in it;
        - if `t` is `typing.Dict` or `typing.Mapping`, check that `obj` is a dict and return a dict with recursively JSON-decoded keys and values from `obj` (first parsing the keys from strings in all those cases where `typing_json.encoding.to_json_obj` would have stringified them);
        - if `t` is `typing.OrderedDict`, check that `obj` is a `collections.OrderedDict` and return a `collections.OrderedDict` with recursively JSON-decoded keys and values from `obj` (first parsing the keys from strings in all those cases where `typing_json.encoding.to_json_obj` would have stringified them);

        If `t` is a namedtuple (according to `typing_json.typechecking.is_namedtuple`), `obj` must be a dictionary (not necessarily ordered, although namedtuple are JSON-encoded as such).
        The keys for the dictionary must form a subset of all field names for the namedtuple `t`, including at least all names of fields without default value.
        An instance of `t` is then constructed (and returned) by assigning to fields having names in the dictionary the JSON decoding of the corresponding values in the dictionary, and to all other fields the default values specified by `t`.

        As an exception to the above rule, decoding of namedtuples is allowed from lists of values, in the same order as the namedtuple fields they are to be assigned to.
        Missing values are allowed at the end and are filled with default field values.
        No excess values are allowed.
        This is to support the default `json` library behaviour on namedtuples, encoded as lists of field values.

        (Version 0.1.1.post1)
    """
    # pylint: disable = too-many-branches, too-many-statements, too-many-return-statements
    trace: List[str] = []
    def failure_callback(message: str) -> None:
        trace.append(message)
    if not is_json_encodable(t, failure_callback=failure_callback):
        # Argument `t` must be JSON encodable.
        raise TypeError("Type %s is not json-encodable. Trace:\n%s"%(str(t), "\n".join(trace)))
    if t in JSON_BASE_TYPES:
        # JSON basic types are returned unaltered, with the exception of casting `Decimal` to `int`/`float` if `cast_decimal` is `True`.
        if t == int and cast_decimal and isinstance(obj, Decimal) and obj == obj.to_integral_value():
            return int(obj)
        if t == float and cast_decimal and isinstance(obj, Decimal):
            return float(obj)
        if t == float and isinstance(obj, int) and obj is not True and obj is not False:
            return float(obj)
        if not is_instance(obj, t, cast_decimal=cast_decimal):
            raise TypeError("Object %s is not of json basic type t=%s."%(short_str(obj), str(t)))
        return obj
    if t in (None, type(None)):
        # The only value of `NoneType` is `None`, which is returned unaltered.
        if obj is not None:
            raise TypeError("Object %s is not None (t=%s)."%(short_str(obj), str(t)))
        return None
    if t == Decimal:
        # Instances of `decimal.Decimal` are decoded from `int` or `string`, as well as from `float` if `cast_decimal` is `True`
        try:
            if isinstance(obj, (int, str, Decimal)) and obj is not True and obj is not False:
                return Decimal(obj)
            if cast_decimal and isinstance(obj, float) and obj is not True and obj is not False:
                return Decimal(obj)
        except InvalidOperation:
            ...
        raise TypeError("Object %s is not decimal.Decimal (t=%s)."%(short_str(obj), str(t)))
    if isinstance(t, EnumMeta):
        # For enumerations, use the `t.__members__` dictionary to convert the string name into an enumeration value.
        if not isinstance(obj, str):
            raise TypeError("Object %s is not a string (t=%s)."%(short_str(obj), str(t)))
        if obj not in t.__members__: # type: ignore
            raise TypeError("Object %s is not the string of a value of the enum (t=%s)."%(short_str(obj), str(t)))
        return t.__members__[obj] # type: ignore # pylint:disable=protected-access
    if is_namedtuple(t):
        fields = getattr(t, "_fields")
        field_types = getattr(t, "_field_types")
        field_defaults = getattr(t, "_field_defaults")
        if isinstance(obj, list):
            # there is a special provision for decoding namedtuples from lists (the standard encoding done by the builtin json library)
            if len(fields) < len(obj):
                raise TypeError("Object %s provides too many values for namedtuple type t=%s."%(short_str(obj), str(t)))
            return_val = t(*tuple(from_json_obj(obj[i] if i < len(obj) else field_defaults[field], field_types[field], cast_decimal=cast_decimal) for i, field in enumerate(fields)))
            assert is_instance(return_val, t, cast_decimal=cast_decimal)
            return return_val
        if not isinstance(obj, (dict, OrderedDict)):
            # Namedtuples are ordinarily decoded from dictionaries, not necessarily ordered (though they are encoded as ordered dictionaries).
            raise TypeError("Object %s is not (ordered) dictionary (t=%s)."%(short_str(obj), str(t))) # pylint:disable=line-too-long
        converted_dict: OrderedDict() = {} # type:ignore
        if set(obj.keys()).union(set(field_defaults.keys())) != set(field_types.keys()):
            # raise an error if the keys provided by the object together with the names of fields with default values don't yield exactly the names of all fields for the namedtuple
            key_diff = set(obj.keys()).union(set(field_defaults.keys())) - set(field_types.keys())
            if key_diff:
                raise TypeError("Object %s does not have the required keys: t=%s, extra keys %s."%(short_str(obj), str(t), str(key_diff))) # pylint:disable=line-too-long
            key_diff = set(field_types.keys()) - set(obj.keys()).union(set(field_defaults.keys()))
            raise TypeError("Object %s does not have the required keys: t=%s, missing keys %s."%(short_str(obj), str(t), str(key_diff))) # pylint:disable=line-too-long
        for field in fields:
            field_type = field_types[field]
            if not field in obj:
                # for each field not appearing in the JSON object, use the field default value
                converted_dict[field] = field_defaults[field]
            else:
                # for each field appearing in the JSON object, decoding the corresponding value
                converted_dict[field] = from_json_obj(obj[field], field_type)
        return_val = t(**converted_dict)
        assert is_instance(return_val, t, cast_decimal=cast_decimal)
        return return_val
    if hasattr(t, "__origin__") and hasattr(t, "__args__"):
        # `typing` generics
        if t.__origin__ is Union:
            # For `typing.Union` (and `typing.Optional`), attempt to decode the value using the generic type arguments in sequence
            for s in t.__args__:
                try:
                    return_val = from_json_obj(obj, s)
                    assert is_instance(return_val, t, cast_decimal=cast_decimal)
                    return return_val
                except TypeError:
                    continue
            raise TypeError("Object %s is not convertible to any of the types in %s."%(short_str(obj), str(t)))
        if t.__origin__ is Literal:
            # for `typing_extensions.Literal`, check that the object is an instance of `t` and then return it unaltered
            trace = []
            if not is_instance(obj, t, failure_callback=failure_callback, cast_decimal=cast_decimal):
                raise TypeError("Object %s is not allowed (t=%s). Trace:\n%s"%(short_str(obj), str(t), "\n".join(trace)))
            return obj
        if t.__origin__ is list:
            # for `typing.List`, expect a list and return a list with recursively JSON-decoded elements
            if not isinstance(obj, list):
                raise TypeError("Object %s is not list (t=%s)."%(short_str(obj), str(t)))
            return_val = list(from_json_obj(x, t.__args__[0]) for x in obj)
            assert is_instance(return_val, t, cast_decimal=cast_decimal)
            return return_val
        if t.__origin__ is deque:
            # for `typing.Deque`, expect a list and return a deque with recursively JSON-decoded elements
            if not isinstance(obj, list):
                raise TypeError("Object %s is not list (t=%s)."%(short_str(obj), str(t)))
            return_val = deque(from_json_obj(x, t.__args__[0]) for x in obj)
            assert is_instance(return_val, t, cast_decimal=cast_decimal)
            return return_val
        if t.__origin__ is set:
            # for `typing.Set`, expect a list and return a set with recursively JSON-decoded elements
            if not isinstance(obj, list):
                raise TypeError("Object %s is not list (t=%s)."%(short_str(obj), str(t)))
            return_val = set(from_json_obj(x, t.__args__[0]) for x in obj)
            assert is_instance(return_val, t, cast_decimal=cast_decimal)
            return return_val
        if t.__origin__ is frozenset:
            # for `typing.FrozenSet`, expect a list and return a frozenset with recursively JSON-decoded elements
            if not isinstance(obj, list):
                raise TypeError("Object %s is not list (t=%s)."%(short_str(obj), str(t)))
            return_val = frozenset(from_json_obj(x, t.__args__[0]) for x in obj)
            assert is_instance(return_val, t, cast_decimal=cast_decimal)
            return return_val
        if t.__origin__ is tuple:
            # for `typing.Tuple`, expect a list and return a tuple with recursively JSON-decoded elements
            if not isinstance(obj, list):
                raise TypeError("Object %s is not list (t=%s)."%(short_str(obj), str(t)))
            if len(t.__args__) == 2 and t.__args__[1] is ...: # pylint:disable=no-else-return
                return_val = tuple(from_json_obj(x, t.__args__[0]) for x in obj)
                assert is_instance(return_val, t, cast_decimal=cast_decimal)
                return return_val
            else:
                if len(obj) != len(t.__args__):
                    raise TypeError("List %s is of incorrect length (t=%s)."%(short_str(obj), str(t)))
                return_val = tuple(from_json_obj(x, t.__args__[i]) for i, x in enumerate(obj))
                assert is_instance(return_val, t, cast_decimal=cast_decimal)
                return return_val
        if t.__origin__ in (dict, Mapping):
            # for `typing.Dict` and `typing.Mapping`, expect a dict and return a dict with recursively JSON-decoded values and keys (parsing keys from strings in all those cases where they would have been stringified)
            if not isinstance(obj, (dict, OrderedDict)):
                raise TypeError("Object %s is not dict or OrderedDict (t=%s)."%(short_str(obj), str(t)))
            converted_dict = dict() # type:ignore
            for field in obj:
                if t.__args__[0] in JSON_BASE_TYPES:
                    if not is_instance(field, t.__args__[0], cast_decimal=cast_decimal):
                        raise TypeError("Object key %s is not of json basic type %s (t=%s)."%(field, str(t.__args__[0]), str(t)))
                    converted_field = field
                elif isinstance(t.__args__[0], EnumMeta) or hasattr(t.__args__[0], "__origin__") and t.__args__[0].__origin__ is Literal:
                    converted_field = from_json_obj(field, t.__args__[0])
                else:
                    converted_field = from_json_obj(json.loads(field), t.__args__[0])
                converted_dict[converted_field] = from_json_obj(obj[field], t.__args__[1])
            assert is_instance(converted_dict, t, cast_decimal=cast_decimal)
            return converted_dict
        if t.__origin__ is OrderedDict:
            # for `typing.OrderedDict`, expect a `collections.OrderedDict` and return an ordered dict with recursively JSON-decoded values and keys (parsing keys from strings in all those cases where they would have been stringified)
            if not isinstance(obj, OrderedDict):
                raise TypeError("Object %s is not OrderedDict (t=%s)."%(short_str(obj), str(t)))
            converted_dict = OrderedDict() # type:ignore
            for field in obj:
                if t.__args__[0] in JSON_BASE_TYPES:
                    if not isinstance(field, t.__args__[0]):
                        raise TypeError("Object key %s not of json basic type %s (t=%s)."%(field, str(t.__args__[0]), str(t)))
                    converted_field = field
                elif isinstance(t.__args__[0], EnumMeta) or hasattr(t.__args__[0], "__origin__") and t.__args__[0].__origin__ is Literal:
                    converted_field = from_json_obj(field, t.__args__[0])
                else:
                    converted_field = from_json_obj(json.loads(field), t.__args__[0])
                converted_dict[converted_field] = from_json_obj(obj[field], t.__args__[1])
            assert is_instance(converted_dict, t, cast_decimal=cast_decimal)
            return converted_dict
    raise AssertionError(_UNREACHABLE_ERROR_MSG) # pragma: no cover



""" setup.py created according to https://packaging.python.org/tutorials/packaging-projects """

import setuptools #type:ignore

with open("README.md", "r") as fh:
    long_description: str = fh.read()

setuptools.setup(
    name="typing-json",
    version="0.1.1.post1",
    author="sg495",
    author_email="sg495@users.noreply.github.com",
    description="Type-aware Python JSON serialization and validation.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sg495/typing-json",
    packages=setuptools.find_packages(exclude=["test"]),
    classifiers=[ # see https://pypi.org/classifiers/
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Natural Language :: English",
        "Typing :: Typed",
    ],
    package_data={"": ["LICENSE", "README.md"],
                  "typing_json": ["typing_json/py.typed"],
                 },
    include_package_data=True,
    install_requires=[
        "typing_extensions",
    ],
)



#pylint:disable = line-too-long, invalid-name
"""
    The `typing_json.typechecking` module provides functionality for dynamic typechecking.

    The core functionality is provided by `typing_json.typechecking.is_instance`, which extends
    the builtin `isinstance` to deal with certain typed collections created using the `typing` module,
    as well as literal types, optional types, unions and (certain) typed namedtuples.

    (Version: 0.1.0)
"""

# standard imports
from collections import deque, OrderedDict
from collections.abc import Mapping
from decimal import Decimal
from enum import EnumMeta
import textwrap
from typing import Any, Callable, Optional, Tuple, Type, Union

# external dependencies
from typing_extensions import Literal


JSON_BASE_TYPES: Tuple[type, ...] = (bool, int, float, str, type(None))
""" Base types for JSON. """


KEYABLE_BASE_TYPES = (bool, int, float, Decimal, complex, str, bytes, range, type)
""" Base types that can be used for dictionary keys. """


TYPECHECKABLE_BASE_TYPES = (bool, int, float, Decimal, complex, str, bytes, bytearray, memoryview, list, tuple, range, slice, set, frozenset, dict, type, deque, OrderedDict, object)
""" Base types that can be typechecked. """


_UNREACHABLE_ERROR_MSG = "Should never reach this point, please open an issue on GitHub."


def _not_keyable(message: str, failure_callback: Optional[Callable[[str], None]]) -> Literal[False]:
    """ Utility message to fail (return `False`) by first calling an optional failure callback. """
    if failure_callback:
        failure_callback(message)
    return False


def is_keyable(t: Type, failure_callback: Optional[Callable[[str], None]] = None) -> bool:
    """
        Check whether `t` is a type that can be used as a key when encoding/decoding mappings
        using this library.
        This function is used only in `typing_json.encoding.is_json_encodable`, to decided whether a
        mapping type is JSON encodable using this library.

        The optional parameter `failure_callback` can be used to collect a detailed trace of
        the reasons behind this method returning `False` on a given type `t`.

        At present, a type is keyable if it satisfies one of the following conditions:

        - it is one of `bool`, `int`, `float`, `decimal.Decimal`, `complex`, `str`, `bytes`, `range`, `type`;
        - it is `None` or an enumeration (i.e. `isinstance(t, EnumMeta)`);
        - it is a variadic `typing.Tuple`, a `typing.FrozenSet` or a `typing.Optional` and its generic type argument is keyable;
        - it is a fixed-length `typing.Tuple` or a `typing.Union` and all of its generic type arguments are keyable;
        - it is a `typing_extensions.Literal`;
        - it is a named tuple according to `typing_json.typechecking.is_namedtuple` and all of its fields are of keyable type.

    """
    # pylint: disable = too-many-return-statements, too-many-branches
    if t in KEYABLE_BASE_TYPES:
        # Types in the `KEYABLE_BASE_TYPES` collection are keyable.
        return True
    if t in (None, type(None)):
        # `None` is keyable.
        return True
    if isinstance(t, EnumMeta):
        # Enum types are keyable.
        return True
    if hasattr(t, "__origin__") and hasattr(t, "__args__"):
        # Parametric types in the `typing` module.
        if t.__origin__ in (frozenset, Union, Optional):
            # The types `typing.FrozenSet`, `typing.Union` and `typing.Optional` are keyable
            # if all of their type arguments are keyable.
            if all(is_keyable(s, failure_callback=failure_callback) for s in t.__args__):
                return True
            return _not_keyable("Not all type arguments of type %s are keyable."%str(t), failure_callback=failure_callback)
        if t.__origin__ is tuple:
            # The type `typing.Tuple` is keyable if all of its type arguments are keyable.
            if len(t.__args__) == 2 and t.__args__[1] == ...:
                # This is the case of variadic `typing.Tuple`.
                if is_keyable(t.__args__[0], failure_callback=failure_callback):
                    return True
            else:
                # This is the case of fixed-length `typing.Tuple`.
                if all(is_keyable(s, failure_callback=failure_callback) for s in t.__args__):
                    return True
            return _not_keyable("Not all type arguments of type %s are keyable."%str(t), failure_callback=failure_callback)
        if t.__origin__ is Literal:
            # The type `typing_extensions.Literal` is keyable if all of its type arguments are keyable.
            # Currently, there is no way for this not to be the case, so this always returns true.
            if all(isinstance(s, KEYABLE_BASE_TYPES) or s is None for s in t.__args__):
                return True
            raise AssertionError(_UNREACHABLE_ERROR_MSG) # pragma: no cover
    if is_namedtuple(t, check_keyable=True):
        # Types inheriting from `typing.NamedTuple` are keyable if all their fields have keyable type.
        return True
    return _not_keyable("Type %s is not keyable."%str(t), failure_callback=failure_callback)


def _not_typecheckable(message: str, failure_callback: Optional[Callable[[str], None]]) -> Literal[False]:
    """ Utility message to fail (return `False`) by first calling an optional failure callback. """
    if failure_callback:
        failure_callback(message)
    return False


def is_typecheckable(t: Any, failure_callback: Optional[Callable[[str], None]] = None) -> bool:
    """
        Checks whether `t` can be type-checked according to the `typing_json` library.
        This is a pre-requisite for `t` to be JSON encodable/decodable in the library.
        It is also a pre-requisite for all fields of types deemed to be namedtuples
        by `typing_json.typechecking.is_namedtuple`.

        The optional parameter `failure_callback` can be used to collect a detailed trace of
        the reasons behind this method returning `False` on a given type `t`.

        At present, a type is typecheckable if it satisfies one of the following conditions:

        - it is one of the basic typecheckable types: `bool`, `int`, `float`, `decimal.Decimal`, `complex`, `str`, `bytes`, `bytearray`, `memoryview`, `list`, `tuple`, `range`, `slice`, `set`, `frozenset`, `dict`, `type`, `collections.deque`, `collections.OrderedDict`, `object`;
        - it is `None`, `typing.Any` or an enumeration (i.e. `isinstance(t, EnumMeta)`);
        - it is one of `typing.List`, `typing.Set`, `typing.FrozenSet`, `typing.Deque`, `typing.Optional` or variadic `typing.Tuple` and its generic type argument is typecheckable;
        - it is one of `typing.Dict`, `typing.OrderedDict`, `typing.Mapping`, `typing.Union` or fixed-length `typing.Tuple` and all of its generic type arguments are typecheckable;
        - it is a `typing.Literal` including literals of one of the JSON basic types `bool`, `int`, `float`, `str` or `NoneType`;
        - it is a named tuple according to `typing_json.typechecking.is_namedtuple` and all of its fields are of typecheckable type.
    """
    # pylint: disable = too-many-return-statements, too-many-branches
    if t in TYPECHECKABLE_BASE_TYPES:
        # Types in the `TYPECHECKABLE_BASE_TYPES` collection are all typecheckable.
        return True
    if t in (None, type(None), Any):
        # `None` and `typing.Any` are typecheckable.
        return True
    if isinstance(t, EnumMeta):
        # Enum types are typecheckable.
        return True
    if hasattr(t, "__origin__") and hasattr(t, "__args__"):
        # Parametric types in the `typing` module.
        if t.__origin__ in (list, set, frozenset, dict,
                            deque, OrderedDict, Union, Optional, Mapping):
            # The types `typing.List`, `typing.Set`, `typing.FrozenSet`, `typing.Dict`,
            # `typing.Deque`, `typing.OrderedDict`, `typing.Union` and `typing.Mapping`
            # are typecheckable if all of their type arguments are typecheckable.
            if all(is_typecheckable(s, failure_callback=failure_callback) for s in t.__args__):
                return True
            return _not_typecheckable("Not all type arguments of type %s are typecheckable."%str(t), failure_callback=failure_callback)
        if t.__origin__ is tuple:
            # The type `typing.Tuple` is typecheckable if all of its type arguments are typecheckable.
            if len(t.__args__) == 2 and t.__args__[1] == ...:
                # This is the case of variadic `typing.Tuple`.
                if is_typecheckable(t.__args__[0], failure_callback=failure_callback):
                    return True
            else:
                # This is the case of fixed-length `typing.Tuple`.
                if all(is_typecheckable(s, failure_callback=failure_callback) for s in t.__args__):
                    return True
            return _not_typecheckable("Not all type arguments of type %s are typecheckable."%str(t), failure_callback=failure_callback)
        if t.__origin__ is Literal:
            # The type `typing_extensions.Literal` is typecheckable if all of its type arguments are of JSON basic type.
            if all(isinstance(s, JSON_BASE_TYPES) for s in t.__args__):
                return True
            return _not_typecheckable("Not all type arguments of literal type %s are of JSON basic type."%str(t), failure_callback=failure_callback)
    if is_namedtuple(t, failure_callback=failure_callback):
        # Types inheriting from `typing.NamedTuple` are typecheckable, because `is_namedtuple` already
        # enforces fields to be of typecheckable type.
        return True
    return _not_typecheckable("Type %s is not typecheckable."%str(t), failure_callback=failure_callback)


def short_str(obj: Any) -> str:
    """ Returns a shortened string representation of `obj`, for use in error messages. """
    if isinstance(obj, str):
        return "\""+obj+"\""
    return textwrap.shorten(repr(obj), width=30, placeholder="...")


def _not_instance(message: str, failure_callback: Optional[Callable[[str], None]]) -> Literal[False]:
    """ Utility message to fail (return `False`) by first calling an optional failure callback. """
    if failure_callback:
        failure_callback(message)
    return False


def is_instance(obj: Any, t: Type, failure_callback: Optional[Callable[[str], None]] = None, cast_decimal: bool = True) -> bool:
    """
        Checks whether an object `obj` is an instance of type `t`, extending the dynamical typechecking capabilities of the
        builtin `isinstance` to some of the `typing` generics and to certain types constructed with `typing.NamedTuple`.
        On the basic typecheckable types (cf.&nbsp;`typing_json.typechecking.is_typecheckable`), it acts as the builtin `isinstance`,
        with tje following exceptions:

        - if the optional parameter `cast_decimal` is set to `True`, instances of `decimal.Decimal` are deemed to be instances of `float` (and `int` if integral) by this function;
        - the boolean literals `True` and `False` are not deemed of type `int` by this function (cf.&nbsp;https://www.python.org/dev/peps/pep-0285/).
        - instances of `int` are deemed to be instances of `float` by this function.

        The optional parameter `failure_callback` can be used to collect a detailed trace of the reasons behind this function returning `False` on a given object `obj` and type `t`.

        The optional parameter `cast_decimal` (default:&nbsp;`True`) can be used to specify that objects of type `decimal.Decimal`
        which encode integers have to be deemed of type `int` or `float`:

        ```python
            >>> from decimal import Decimal
            >>> is_instance(Decimal(1), int, cast_decimal=True)
            True
            >>> is_instance(Decimal(1.1), float, cast_decimal=True)
            True
            >>> is_instance(Decimal(1.1), int, cast_decimal=True)
            False
            >>> is_instance(Decimal(1), int, cast_decimal=False)
            False
            >>> is_instance(Decimal(1.1), float, cast_decimal=False)
            False
        ```

        Literals in `typing_extensions.Literal` can only be of one of the JSON basic types `bool`, `int`, `float`, `str`, `NoneType`.
    """
    # pylint: disable = too-many-return-statements, too-many-branches, too-many-statements
    if t in TYPECHECKABLE_BASE_TYPES:
        # for basic types, use builtin `isinstance`.
        if t == int and (obj is True or obj is False):
            # special case to deal with the fact that `bool` inherits from `int`, see https://www.python.org/dev/peps/pep-0285/
            return False
        if t == int and isinstance(obj, Decimal) and cast_decimal and obj == obj.to_integral_value():
            # special case to deal with `decimal.Decimal` being used to encode integers:
            return True
        if t == float and isinstance(obj, Decimal) and cast_decimal:
            # special case to deal with `decimal.Decimal` being used to encode floats:
            return True
        if t == float and isinstance(obj, int) and obj is not True and obj is not False:
            return True
        if isinstance(obj, t):
            return True
        return _not_instance("Value %s is not of type %s."%(short_str(obj), str(t)), failure_callback=failure_callback)
    if t in (None, type(None)):
        # for `None`, use `is None`.
        if obj is None:
            return True
        return _not_instance("Value %s is not of type %s."%(short_str(obj), str(t)), failure_callback=failure_callback)
    if t == Any:
        # For `typing.Any`, always return `True`.
        return True
    if isinstance(t, EnumMeta):
        # For enums, check whether `obj` is one of the values of the enumeration `t`.
        if obj in t.__members__.values(): # type: ignore
            return True
        return _not_instance("Value %s is not of enum type %s."%(short_str(obj), str(t)), failure_callback=failure_callback)
    if is_namedtuple(t, failure_callback=failure_callback):
        # For namedtuples, check that all fields are defined and have value of designated type.
        if obj.__class__ != t:
            return _not_instance("Value %s is not of type %s: wrong class %s."%(short_str(obj), str(t), str(obj.__class__)), failure_callback=failure_callback)
        field_types = getattr(t, "_field_types")
        for field in field_types:
            if not hasattr(obj, field):
                raise AssertionError(_UNREACHABLE_ERROR_MSG) # pragma: no cover
                # return _not_instance("Value %s is not of type %s: missing field %s."%(short_str(obj), str(t), field), failure_callback=failure_callback)
            field_val = getattr(obj, field)
            if not is_instance(field_val, field_types[field], failure_callback=failure_callback, cast_decimal=cast_decimal):
                return _not_instance("Value %s is not of type %s: wrong type %s for field %s, expected %s."%(short_str(obj), str(t), str(type(field_val)), field, str(field_types[field])), failure_callback=failure_callback)
        return True
    if hasattr(t, "__origin__") and hasattr(t, "__args__"):
        # Special cases for `typing` generics.
        if t.__origin__ is Union: # Union[T1, T2, ..., TN] or Optional[T]
            # For `typing.Union` (including `typing.Optional`), check that `obj` is instance of one of the type parameters of `typing.Union`.
            if any(is_instance(obj, s, failure_callback=failure_callback, cast_decimal=cast_decimal) for s in t.__args__):
                return True
            return _not_instance("Value %s does not match any of the types in %s."%(short_str(obj), str(t)), failure_callback=failure_callback)
        if t.__origin__ is Literal: # Literal[val1, val2, ..., valN]
            # For `typing_extensions.Literal`, check that `obj` equals one of the literals parameters of `typing_extensions.Literal`.
            if any(obj == s for s in t.__args__):
                return True
            return _not_instance("Value %s does not match any of the values in %s."%(short_str(obj), str(t)), failure_callback=failure_callback)
        if t.__origin__ is list: # List[T]
            # For `typing.List`, check that `obj` is a `list` and that all elements of `obj` are instances of the `typing.List` type parameter.
            if not isinstance(obj, list):
                return _not_instance("Value %s is not a list."%short_str(obj), failure_callback=failure_callback)
            if all(is_instance(x, t.__args__[0], failure_callback=failure_callback, cast_decimal=cast_decimal) for x in obj):
                return True
            return _not_instance("Not all elements of %s are of type %s."%(short_str(obj), str(t.__args__[0])), failure_callback=failure_callback)
        if t.__origin__ is tuple: # Tuple[T1, T2, ..., TN] or Tuple[T, ...] (with an actual ellipse `...` as the second type parameter of `typing.Tuple`)
            # For `typing.Tuple`, check that `obj` is a `tuple` and that all elements of `obj` are instances of the `typing.Tuple` type parameter(s).
            if not isinstance(obj, tuple):
                return _not_instance("Value %s is not a tuple."%short_str(obj), failure_callback=failure_callback)
            if len(t.__args__) == 2 and t.__args__[1] is ...: # pylint:disable=no-else-return
                # for variadic tuples, all elements have to be of the same type.
                if all(is_instance(x, t.__args__[0], failure_callback=failure_callback, cast_decimal=cast_decimal) for x in obj):
                    return True
                return _not_instance("Not all elements of %s are of type %s."%(short_str(obj), str(t.__args__[0])), failure_callback=failure_callback)
            else:
                # for fixed-length tuples, each element has to be of the correct positional type.
                if len(obj) != len(t.__args__):
                    return _not_instance("Tuple %s is of the wrong length for type %s"%(short_str(obj), str(t)), failure_callback=failure_callback)
                if all(is_instance(x, t.__args__[i], failure_callback=failure_callback, cast_decimal=cast_decimal) for i, x in enumerate(obj)):
                    return True
                return _not_instance("Not all values in %s are of the respective types specified by %s"%(short_str(obj), str(t)), failure_callback=failure_callback)
        if t.__origin__ is set: # Set[T]
            # For `typing.Set`, check that `obj` is a `set` and that all elements of `obj` are instances of the `typing.Set` type parameter.
            if not isinstance(obj, set):
                return _not_instance("Value %s is not a set."%short_str(obj), failure_callback=failure_callback)
            if all(is_instance(x, t.__args__[0], failure_callback=failure_callback, cast_decimal=cast_decimal) for x in obj):
                return True
            return _not_instance("Not all elements of %s are of type %s."%(short_str(obj), str(t.__args__[0])), failure_callback=failure_callback)
        if t.__origin__ is frozenset: # FrozenSet[T]
            # For `typing.FrozenSet`, check that `obj` is a `frozenset` and that all elements of `obj` are instances of the `typing.FrozenSet` type parameter.
            if not isinstance(obj, frozenset):
                return _not_instance("Value %s is not a frozenset."%short_str(obj), failure_callback=failure_callback)
            if all(is_instance(x, t.__args__[0], failure_callback=failure_callback, cast_decimal=cast_decimal) for x in obj):
                return True
            return _not_instance("Not all elements of %s are of type %s."%(short_str(obj), str(t.__args__[0])), failure_callback=failure_callback)
        if t.__origin__ is deque: # Deque[T]
            # For `typing.Deque`, check that `obj` is a `deque` and that all elements of `obj` are instances of the `typing.Deque` type parameter.
            if not isinstance(obj, deque):
                return _not_instance("Value %s is not a deque."%short_str(obj), failure_callback=failure_callback)
            if all(is_instance(x, t.__args__[0], failure_callback=failure_callback, cast_decimal=cast_decimal) for x in obj):
                return True
            return _not_instance("Not all elements of %s are of type %s."%(short_str(obj), str(t.__args__[0])), failure_callback=failure_callback)
        if t.__origin__ is dict: # Dict[K,V]
            # For `typing.Dict`, check that `obj` is a `dict`,
            # check that all keys of `obj` are instances of the first `typing.Dict` type parameter,
            # and check that all values of `obj` are instances of the econd `typing.Dict` type parameter.
            if not isinstance(obj, (dict)):
                return _not_instance("Value %s is not a dict."%short_str(obj), failure_callback=failure_callback)
            if not all(is_instance(x, t.__args__[0], failure_callback=failure_callback, cast_decimal=cast_decimal) for x in obj):
                return _not_instance("Not all keys of %s are of type %s."%(short_str(obj), str(t.__args__[0])), failure_callback=failure_callback)
            if not all(is_instance(obj[x], t.__args__[1], failure_callback=failure_callback, cast_decimal=cast_decimal) for x in obj):
                return _not_instance("Not all values of %s are of type %s."%(short_str(obj), str(t.__args__[1])), failure_callback=failure_callback)
            return True
        if t.__origin__ is OrderedDict: # OrderedDict[K,V]
            # For `typing.OrderedDict`, check that `obj` is a `collections.OrderedDict`,
            # check that all keys of `obj` are instances of the first `typing.OrderedDict` type parameter,
            # and check that all values of `obj` are instances of the econd `typing.OrderedDict` type parameter.
            if not isinstance(obj, (OrderedDict)):
                return _not_instance("Value %s is not an OrderedDict."%short_str(obj), failure_callback=failure_callback)
            if not all(is_instance(x, t.__args__[0], failure_callback=failure_callback, cast_decimal=cast_decimal) for x in obj):
                return _not_instance("Not all keys of %s are of type %s."%(short_str(obj), str(t.__args__[0])), failure_callback=failure_callback)
            if not all(is_instance(obj[x], t.__args__[1], failure_callback=failure_callback, cast_decimal=cast_decimal) for x in obj):
                return _not_instance("Not all values of %s are of type %s."%(short_str(obj), str(t.__args__[1])), failure_callback=failure_callback)
            return True
        if t.__origin__ is Mapping: # Mapping[K,V], used for read-only dictionaries.
            # For `typing.Mapping`, check that `obj` is either a `dict` or a `collections.OrderedDict`,
            # check that all keys of `obj` are instances of the first `typing.Mapping` type parameter,
            # and check that all values of `obj` are instances of the econd `typing.Mapping` type parameter.
            if not isinstance(obj, (dict, OrderedDict)):
                return _not_instance("Value %s is not a dict or OrderedDict."%short_str(obj), failure_callback=failure_callback)
            if not all(is_instance(x, t.__args__[0], failure_callback=failure_callback, cast_decimal=cast_decimal) for x in obj):
                return _not_instance("Not all keys of %s are of type %s."%(short_str(obj), str(t.__args__[0])), failure_callback=failure_callback)
            if not all(is_instance(obj[x], t.__args__[1], failure_callback=failure_callback, cast_decimal=cast_decimal) for x in obj):
                return _not_instance("Not all values of %s are of type %s."%(short_str(obj), str(t.__args__[1])), failure_callback=failure_callback)
            return True
    if failure_callback:
        failure_callback("Type %s is not supported."%str(t))
    raise TypeError("Type %s is not supported."%str(t))


def _not_namedtuple(message: str, failure_callback: Optional[Callable[[str], None]]) -> Literal[False]:
    """ Utility message to fail (return `False`) by first calling an optional failure callback. """
    if failure_callback:
        failure_callback(message)
    return False


def is_namedtuple(t: Type, failure_callback: Optional[Callable[[str], None]] = None, check_typecheckable: bool = True, check_keyable: bool = False, cast_decimal: bool = True) -> bool:
    """
        Checks whether `t` is a type constructed using `typing.NamedTuple`, using the following procedure:

        1. checks for existence of the attribute `t.__bases__`, containing the base classes of `t`;
        2. checks that there is exactly one base class in `t.__bases__`, namely `tuple`;
        3. checks for existence of the attribute `t._fields`, containing the fields names of`t`;
        4. checks that `t._fields` is a tuple of strings.
        5. checks for existence of the attribute `t._field_types`, containing the field types for `t`;
        6. checks that `t._field_tyes` is a dictionary with exactly the elements of `t._fields` as its keys;
        7. checks for the existence of the attribute `t._field_defaults`, containing the default values for fields of `t`;
        8. checks that `t._field_defaults` is a dictionary, and that all of its keys (if any) appear in `t._fields`;
        9. checks that all fields in `t._fields` also appear in `dir(t)`.

        The procedure above weeds out many incorrect examples, but certainly needs to be improved to catch all exceptions.

        If the optional parameter `check_typecheckable` is set to `True` (default: `True`), an additional check is added between 6. and 7. above
        for all field types to be typecheckable according to `typing_json.typechecking.is_typecheckable`.
        If the optional parameter `check_keyable` is set to `True` (default: `False`), an additional check is added between 6. and 7. above
        for all field types to be keyable according to `typing_json.typechecking.is_keyable`.
        If the optional parameter `check_typecheckable` is set to `True` (default: `True`), an additional check is added between 8. and 9. above
        for all field default values to be instances of the corresponding field types, according to `typing_json.typechecking.is_instance` (the value
        of the optional parameter `cast_decimal` is passed to `typing_json.typechecking.is_instance` when performing this check).
    """
    # pylint:disable = too-many-return-statements, too-many-branches, protected-access
    if not hasattr(t, "__bases__"):
        return _not_namedtuple("Type %s has no attribute __bases__."%str(t), failure_callback=failure_callback)
    base_classes = t.__bases__
    if len(base_classes) != 1 or base_classes[0] != tuple:
        return _not_namedtuple("Attribute bases for type %s should be [tuple], found %s instead"%(str(t), str(t.__bases__)), failure_callback=failure_callback)
    if not hasattr(t, "_fields"):
        return _not_namedtuple("Type %s has no attribute _fields."%str(t), failure_callback=failure_callback)
    fields = getattr(t, "_fields")
    if not isinstance(fields, tuple):
        return _not_namedtuple("Attribute _fields for type %s should be a tuple, found %s instead."%(str(t), str(fields)), failure_callback=failure_callback)
    if not all(isinstance(n, str) for n in fields):
        return _not_namedtuple("Attribute _fields for type %s should be a tuple of strings, found %s instead."%(str(t), str(fields)), failure_callback=failure_callback)
    if not hasattr(t, "_field_types"):
        return _not_namedtuple("Type %s has no attribute _field_types."%str(t), failure_callback=failure_callback)
    field_types = getattr(t, "_field_types", None)
    if not isinstance(field_types, dict):
        return _not_namedtuple("Attribute _field_types for type %s should be a dict, found %s instead."%(str(t), str(field_types)), failure_callback=failure_callback)
    for n in fields:
        if not n in field_types:
            return _not_namedtuple("Field %s appears in _fields but not in _field_types for type %s."%(n, str(t)), failure_callback=failure_callback)
    for n in field_types:
        if not n in fields:
            return _not_namedtuple("Field %s appears in _field_types but not in _fields for type %s."%(n, str(t)), failure_callback=failure_callback)
        if check_typecheckable and not is_typecheckable(field_types[n], failure_callback=failure_callback):
            return _not_namedtuple("Field %s for type %s has non-typecheckable field type %s."%(n, str(t), str(field_types[n])), failure_callback=failure_callback)
        if check_keyable and not is_keyable(field_types[n], failure_callback=failure_callback):
            return _not_namedtuple("Field %s for type %s has non-keyable field type %s."%(n, str(t), str(field_types[n])), failure_callback=failure_callback)
    if not hasattr(t, "_field_defaults"):
        return _not_namedtuple("Type %s has no attribute _field_defaults."%str(t), failure_callback=failure_callback)
    field_defaults = getattr(t, "_field_defaults")
    if not isinstance(field_defaults, dict):
        return _not_namedtuple("Attribute _field_types for type %s should be a dict, found %s instead."%(str(t), str(field_types)), failure_callback=failure_callback)
    for n in field_defaults:
        if not n in fields:
            return _not_namedtuple("Field %s appears in _field_defaults but not in _fields for type %s."%(n, str(t)), failure_callback=failure_callback)
        if check_typecheckable and not is_instance(field_defaults[n], field_types[n], failure_callback=failure_callback, cast_decimal=cast_decimal):
            return _not_namedtuple("Default value for field %s of type %s should be of type %s, found type %s instead."%(n, str(t), str(field_types[n]), str(type(field_defaults[n]))), failure_callback=failure_callback)
    for n in fields:
        if n not in dir(t):
            return _not_namedtuple("Field %s appears in _fields but not in dir(%s)."%(n, str(t)), failure_callback=failure_callback)
    return True
